# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import psutil
import appdirs
import threading
import importlib

import parsl
import tomlkit

PARALLEL_CONFIG = threading.local()
PARALLEL_CONFIG.dfk = None
PARALLEL_CONFIG.parallel_config = None
PARALLEL_CONFIG.action_executor_mapping = {}

# We write a default config to a location in the conda env if there is an
# active conda env. If there is not an active conda env (most likely because we
# are using Docker) then the path we want to write the default to will not
# exist, so we will not write a default, we will just load it from memory
CONDA_PREFIX = os.environ.get('CONDA_PREFIX', '')
VENDORED_FP = os.path.join(CONDA_PREFIX, 'etc', 'qiime2_config.toml')

VENDORED_CONFIG = {
    'parsl': {
        'strategy': 'None',
        'executors': [
            {'class': 'ThreadPoolExecutor', 'label': 'default',
                'max_threads': max(psutil.cpu_count() - 1, 1)},
            {'class': 'HighThroughputExecutor', 'label': 'htex',
                'max_workers': max(psutil.cpu_count() - 1, 1),
                'provider': {'class': 'LocalProvider'}}
            ]
        }
    }

# As near as I can tell, loading a config with a HighThroughputExecutor leaks
# open sockets. This leads to issues (especially on osx) with "too many open
# files" errors while running the test, so this test config with no
# HighThroughputExecutor was created to mitigate that scenario. This config is
# only to be used in tests that do not specifically need to test multiple
# different executors
_TEST_CONFIG_ = {
    'parsl': {
        'strategy': 'None',
        'executors': [
            {'class': 'ThreadPoolExecutor', 'label': 'default',
                'max_threads': 1},
            {'class': '_TEST_EXECUTOR_', 'label': 'test',
                'max_threads': 1}
            ]
        }
    }

# Directs keys in the config whose values need to be objects to the module that
# contains the class they need to instantiate
PARSL_CHANNEL = 'parsl.channels'
PARSL_DATA_PROVIDER = 'parsl.data_provider'
PARSL_EXECUTOR = 'parsl.executors'
PARSL_LAUNCHER = 'parsl.launchers'
PARSL_MONITORING = 'parsl.monitoring'
PARSL_PROVIDER = 'parsl.providers'
module_paths = {
    'channel': PARSL_CHANNEL,
    'channels': PARSL_CHANNEL,
    'data_provider': PARSL_DATA_PROVIDER,
    'data_providers': PARSL_DATA_PROVIDER,
    'executor': PARSL_EXECUTOR,
    'executors': PARSL_EXECUTOR,
    'launcher': PARSL_LAUNCHER,
    'launchers': PARSL_LAUNCHER,
    'monitoring': PARSL_MONITORING,
    'provider': PARSL_PROVIDER,
    'providers': PARSL_PROVIDER
}


def _setup_parallel():
    """Sets the parsl config and action executor mapping to the values set on
    the thread local
    """
    parallel_config = PARALLEL_CONFIG.parallel_config
    mapping = PARALLEL_CONFIG.action_executor_mapping

    # If we are running tests, get the test config not the normal one
    if os.environ.get('QIIMETEST') is not None and parallel_config is None:
        _parallel_config, _mapping = get_config_from_dict(_TEST_CONFIG_)
    else:
        # If they did not already supply a config, we get the vendored one
        if parallel_config is None:
            config_fp = _get_vendored_config()

            # If we are not in a conda environment, we may not get an fp back
            # (because the vendored fp uses the conda prefix), so we load from
            # the vendored dict. Otherwise we load from the vendored file
            if config_fp is not None:
                _parallel_config, _mapping = get_config_from_file(config_fp)
            else:
                _parallel_config, _mapping = \
                    get_config_from_dict(VENDORED_CONFIG)

    # If they did not supply a parallel_config, set the vendored one
    if parallel_config is None:
        PARALLEL_CONFIG.parallel_config = _parallel_config

    # If they did not supply a mapping, set the vendored one
    if mapping is {}:
        PARALLEL_CONFIG.action_executor_mapping = _mapping

    PARALLEL_CONFIG.dfk = parsl.load(PARALLEL_CONFIG.parallel_config)


def _cleanup_parallel():
    """Ask parsl to cleanup and then remove the currently active dfk
    """
    PARALLEL_CONFIG.dfk.cleanup()
    parsl.clear()


def get_config_from_file(config_fp):
    """Takes a config filepath and determines if the file exists and if so if
    it contains parsl config info.
    """
    with open(config_fp, 'r') as fh:
        # After parsing the file tomlkit has the data wrapped in its own
        # proprietary classes. Unwrap recursively turns these classes into
        # Python built-ins
        #
        # ex: tomlkit.items.Int -> int
        #
        # issue caused by this wrapping
        # https://github.com/Parsl/parsl/issues/3027
        config_dict = tomlkit.load(fh).unwrap()

    return get_config_from_dict(config_dict)


def get_config_from_dict(config_dict):
    parallel_config_dict = config_dict.get('parsl')
    mapping = parallel_config_dict.pop('executor_mapping', {})

    processed_parallel_config_dict = _process_config(parallel_config_dict)

    if processed_parallel_config_dict != {}:
        parallel_config = parsl.Config(**processed_parallel_config_dict)
    else:
        parallel_config = None

    return parallel_config, mapping


def _get_vendored_config():
    # 1. Check envvar
    config_fp = os.environ.get('QIIME2_CONFIG')

    if config_fp is None:
        # 2. Check in user writable location
        # appdirs.user_config_dir(appname='qiime2', author='...')
        if os.path.exists(fp_ := os.path.join(
                appdirs.user_config_dir('qiime2'), 'qiime2_config.toml')):
            config_fp = fp_
        # 3. Check in admin writable location
        # /etc/
        # site_config_dir
        # appdirs.site_config_dir(appname='qiime2, author='...')
        elif os.path.exists(fp_ := os.path.join(
                appdirs.site_config_dir('qiime2'), 'qiime2_config.toml')):
            config_fp = fp_
        # NOTE: These next two are dependent on us being in a conda environment
        # 4. Check in conda env
        # ~/miniconda3/env/{env_name}/conf
        elif CONDA_PREFIX != '' and os.path.exists(fp_ := VENDORED_FP):
            config_fp = fp_
        # 5. Write the vendored config to the vendored location and use
        # that
        elif CONDA_PREFIX != '':
            with open(VENDORED_FP, 'w') as fh:
                tomlkit.dump(VENDORED_CONFIG, fh)
            config_fp = VENDORED_FP

    return config_fp


def _process_config(config_dict):
    """Takes a path to a toml file describing a parsl.Config object and parses
    it into a dictionary of kwargs that can be used to instantiate a
    parsl.Config object.
    """
    config_kwargs = {}

    for key, value in config_dict.items():
        # Parsl likes the string 'none' as opposed to None or 'None'
        if isinstance(value, str) and value.lower() == 'none':
            config_kwargs[key] = value.lower()
        # We have a list of values
        elif isinstance(value, list):
            config_kwargs[key] = []
            for item in value:
                config_kwargs[key].append(_process_key(key, item))
        # We have a single value
        else:
            config_kwargs[key] = _process_key(key, config_dict[key])

    return config_kwargs


def _process_key(key, value):
    """Takes a key given in the parsl config file and turns its value into the
    correct data type or class instance to be used in instantiating a
    parsl.Config object.
    """
    # Our key points to a list
    if isinstance(value, list):
        processed_value = []
        for item in value:
            processed_value.append(_process_key(key, item))
        return processed_value
    # Our key needs to point to some object.
    elif key in module_paths:
        # Get the module our class is from
        module = importlib.import_module(module_paths[key])

        _type = value['class']
        if _type == '_TEST_EXECUTOR_':
            # Only used for tests
            cls = _TEST_EXECUTOR_
        else:
            # Get the class we need to instantiate
            cls = getattr(module, value['class'])

        # Get the kwargs we need to pass to the class constructor
        kwargs = {}
        for k, v in value.items():
            # We already handled this key
            if k != 'class':
                kwargs[k] = _process_key(k, v)

        # Instantiate the class
        return cls(**kwargs)
    # Our key points to primitive data
    else:
        return value


class ParallelConfig():
    def __init__(self, parallel_config=None, action_executor_mapping={}):
        """Tell QIIME 2 how to parsl from the Python API

        action_executor_mapping: maps actions to executors. All unmapped
        actions will be run on the default executor. We check which executor a
        given action is supposed to use when we get ready to run the action, so
        errors will only occur if an action that is being run in a given
        QIIME 2 invocation has been mapped to an executor that does not exist

        parallel_config: Specifies which executors should be created and how
        they should be created. If this is None, it will use the default
        config.
        """
        self.parallel_config = parallel_config
        self.action_executor_mapping = action_executor_mapping

    def __enter__(self):
        """Set this to be our Parsl config on the current thread local
        """
        if PARALLEL_CONFIG.parallel_config is not None:
            raise ValueError('ParallelConfig already loaded, cannot nest '
                             'ParallelConfigs')

        PARALLEL_CONFIG.parallel_config = self.parallel_config
        PARALLEL_CONFIG.action_executor_mapping = self.action_executor_mapping

        _setup_parallel()

    def __exit__(self, *args):
        """Set our Parsl config back to whatever it was before this one
        """
        _cleanup_parallel()

        PARALLEL_CONFIG.dfk = None
        PARALLEL_CONFIG.parallel_config = None
        PARALLEL_CONFIG.action_executor_mapping = {}


def _check_env(cls):
    if 'QIIMETEST' not in os.environ:
        raise ValueError(
            f"Do not instantiate the class '{cls}' when not testing")


class _MASK_CONDA_ENV_():
    """Used to test config loading behavior when outside of a conda environment
    """
    def __init__(self):
        _check_env(self.__class__)

    def __enter__(self):
        global CONDA_PREFIX, VENDORED_FP

        self.old_prefix = CONDA_PREFIX
        self.old_fp = VENDORED_FP

        CONDA_PREFIX = ''
        VENDORED_FP = None

    def __exit__(self, *args):
        global CONDA_PREFIX, VENDORED_FP

        CONDA_PREFIX = self.old_prefix
        VENDORED_FP = self.old_fp


class _TEST_EXECUTOR_(parsl.executors.threads.ThreadPoolExecutor):
    """We needed multiple kinds of executor to ensure we were mapping things
    correctly, but the HighThroughputExecutor was leaking sockets, so we avoid
    creating those during the tests because so many sockets were being opened
    that we were getting "Too many open files" errors, so this gets used as the
    second executor type."""

    def __init__(self, *args, **kwargs):
        _check_env(self.__class__)
        super(_TEST_EXECUTOR_, self).__init__(*args, **kwargs)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import psutil

from qiime2.core.transform import ModelType


def transform(data, *, from_type=None, to_type):
    from_type = type(data) if from_type is None else from_type

    from_model_type = ModelType.from_view_type(from_type)
    to_model_type = ModelType.from_view_type(to_type)
    transformation = from_model_type.make_transformation(to_model_type)

    return transformation(data)


def get_available_cores(n_less: int = 0):
    '''
    Finds the number of currently available (logical) cores. Useful for plugins
    that need to convert a 0 to a concrete number of cores when 0 is not
    supported by the underlying/called software.

    Parameters
    ----------
    n_less : int
        The number of cores less than the total number available to request.
        For example `get_available_cores(n_less=2) with 10 available cores
        will return 8.

    Returns
    -------
    int
        The number of cores to be requested.
    '''
    cpus = psutil.cpu_count()
    if cpus is not None:
        return cpus - n_less

    return 1



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
from enum import IntEnum
import pathlib
import warnings
from typing import Optional, Tuple
from zipfile import ZipFile

from qiime2.core.util import md5sum_directory_zip, from_checksum_format
from qiime2.core.archive.archiver import ChecksumDiff

from .util import get_root_uuid, parse_version


class ValidationCode(IntEnum):
    '''
    Codes indicating the level of validation a ProvDAG has passed.

    The code that determines which ValidationCode an archive receives is by
    necessity scattered.

    INVALID:
        One or more files are known to be missing or unparseable. Occurs
        either when checksum validation fails, or when expected files are
        absent or unparseable.
    VALIDATION_OPTOUT:
        The user opted out of checksum validation. This will be overridden by
        INVALID iff a required file is missing. In this context,
        `checksums.md5` is not required. If data files, for example, have been
        manually modified, the code will remain VALIDATION_OPTOUT, but if an
        action.yaml file is missing, INVALID will result.
    PREDATES_CHECKSUMS:
        The archive format predates the creation of checksums.md5, so full
        validation is impossible. We initially assume validity. This will be
        overridden by INVALID iff an expected file is missing or unparseable.
        If data files, for example, have been manually modified, the code will
        remain PREDATES_CHECKSUMS.
    VALID:
        The archive has passed checksum validation and is "known" to be
        valid. Md5 checksums are technically falsifiable, so this is not a
        guarantee of correctness/authenticity.
    '''
    INVALID = 0
    VALIDATION_OPTOUT = 1
    PREDATES_CHECKSUMS = 2
    VALID = 3


def validate_checksums(
    zf: ZipFile
) -> Tuple[ValidationCode, Optional[ChecksumDiff]]:
    '''
    Uses diff_checksums to validate the archive's provenance, warning the user
    if checksums.md5 is missing, or if the archive is corrupt or has been
    modified.

    Parameters
    ----------
    zf : ZipFile
        The zipfile object of the archive.

    Returns
    -------
    tuple of (ValidationCode, ChecksumDiff)
        If the checksums.md5 fle isn't present set ChecksumDiff to None
        and ValidationCode to INVALID and return.
    '''
    checksum_diff: Optional[ChecksumDiff]
    provenance_is_valid = ValidationCode.VALID

    for fp in zf.namelist():
        if 'checksums.md5' in fp:
            break
    else:
        warnings.warn(
            'The checksums.md5 file is missing from the archive. '
            'Archive may be corrupt or provenance may be false.',
            UserWarning
        )
        return ValidationCode.INVALID, None

    checksum_diff = diff_checksums(zf)
    if checksum_diff != ChecksumDiff({}, {}, {}):
        root_uuid = get_root_uuid(zf)
        warnings.warn(
            f'Checksums are invalid for Archive {root_uuid}\n'
            'Archive may be corrupt or provenance may be false.\n'
            f'Files added since archive creation: {checksum_diff.added}\n'
            'Files removed since archive creation: '
            f'{checksum_diff.removed}\n'
            'Files changed since archive creation: '
            f'{checksum_diff.changed}',
            UserWarning
        )
        provenance_is_valid = ValidationCode.INVALID

    return provenance_is_valid, checksum_diff


def diff_checksums(zf: ZipFile) -> ChecksumDiff:
    '''
    Calculates checksums for all files in an archive (except checksums.md5).
    Compares these against the checksums stored in checksums.md5, returning
    a summary ChecksumDiff.

    Parameters
    ----------
    zf : ZipFile
        The zipfile object of the archive.

    Returns
    -------
    ChecksumDiff
        A tuple of three dicts, one each for added, removed, and changed
        files. Keys are filepaths. For the added and removed dicts
        values are the checksum of the added or removed file. For the changed
        dict values are a tuple of (expected checksum, observed checksum).
    '''
    archive_version, _ = parse_version(zf)
    # TODO: don't think this is ever called
    if int(archive_version) < 5:
        return ChecksumDiff({}, {}, {})

    root_dir = pathlib.Path(get_root_uuid(zf))
    checksum_fp = str(root_dir / 'checksums.md5')
    obs = md5sum_directory_zip(zf)

    exp = {}
    for line in zf.open(checksum_fp):
        fp, checksum = from_checksum_format(str(line, 'utf-8'))
        exp[fp] = checksum

    obs_fps = set(obs)
    exp_fps = set(exp)

    added = {fp: obs[fp] for fp in obs_fps - exp_fps}
    removed = {fp: exp[fp] for fp in exp_fps - obs_fps}
    changed = {
        fp: (exp[fp], obs[fp]) for fp in exp_fps & obs_fps
        if exp[fp] != obs[fp]
    }

    return ChecksumDiff(added=added, removed=removed, changed=changed)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import pathlib
import unittest
import zipfile

from .testing_utilities import CustomAssertions, DummyArtifacts
from ..util import get_root_uuid, get_nonroot_uuid


class GetRootUUIDTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_get_root_uuid(self):
        exp_root_uuids = {
            '0': '89af91c0-033d-4e30-8ac4-f29a3b407dc1',
            '1': '5b929500-e4d6-4d3f-8f5f-93fd95d1117d',
            '2': 'e01f0484-40d4-420e-adcf-ca9be58ed1ee',
            '3': 'aa960110-4069-4b7c-97a3-8a768875e515',
            '4': '856502cb-66f2-45aa-a86c-e484cc9bfd57',
            '5': '48af8384-2b0a-4b26-b85c-11b79c0d6ea6',
            '6': '6facaf61-1676-45eb-ada0-d530be678b27',
        }
        for artifact, exp_uuid in zip(
            self.das.all_artifact_versions, exp_root_uuids.values()
        ):
            with zipfile.ZipFile(artifact.filepath) as zfh:
                self.assertEqual(exp_uuid, get_root_uuid(zfh))


class GetNonRootUUIDTests(unittest.TestCase):
    def test_get_nonroot_uuid(self):
        md_example = pathlib.Path(
            'arch_root/provenance/artifacts/uuid123/metadata.yaml')
        action_example = pathlib.Path(
            'arch_root/provenance/artifacts/uuid123/action/action.yaml')
        exp = 'uuid123'

        self.assertEqual(get_nonroot_uuid(md_example), exp)
        self.assertEqual(get_nonroot_uuid(action_example), exp)


class CustomAssertionsTests(CustomAssertions):
    def test_assert_re_appears_only_once(self):
        t = ("Lick an orange. It tastes like an orange.\n"
             "The strawberries taste like strawberries!\n"
             "The snozzberries taste like snozzberries!")
        self.assertREAppearsOnlyOnce(t, 'Lick an orange')
        self.assertREAppearsOnlyOnce(t, 'tastes like')
        with self.assertRaisesRegex(AssertionError, 'Regex.*match.*orange'):
            self.assertREAppearsOnlyOnce(t, 'orange')
        with self.assertRaisesRegex(AssertionError, 'Regex.*taste like'):
            self.assertREAppearsOnlyOnce(t, 'taste like')
        with self.assertRaisesRegex(AssertionError, 'Regex.*snozzberries'):
            self.assertREAppearsOnlyOnce(t, 'snozzberries')
        with self.assertRaisesRegex(AssertionError, 'Regex.*!'):
            self.assertREAppearsOnlyOnce(t, '!')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import shutil
import tempfile
import unittest

from qiime2.sdk.plugin_manager import PluginManager
from qiime2.core.testing.type import IntSequence1

from ..usage_drivers import ReplayPythonUsage
from ..replay import replay_provenance


class ReplayPythonUsageTests(unittest.TestCase):
    def setUp(self):
        self.pm = PluginManager()
        self.dp = self.pm.plugins['dummy-plugin']
        self.tempdir = tempfile.mkdtemp(
            prefix='qiime2-test-usage-drivers-temp-'
        )

        def return_many_ints() -> (list, list, list, list, list, list):
            return ([1, 2, 3], [4, 5, 6], [7], [4, 4], [0], [9, 8])

        self.dp.methods.register_function(
            function=return_many_ints,
            inputs={},
            parameters={},
            outputs=[
                ('ints1', IntSequence1),
                ('ints2', IntSequence1),
                ('ints3', IntSequence1),
                ('ints4', IntSequence1),
                ('ints5', IntSequence1),
                ('ints6', IntSequence1),
            ],
            output_descriptions={
                'ints1': 'ints',
                'ints2': 'ints',
                'ints3': 'ints',
                'ints4': 'ints',
                'ints5': 'ints',
                'ints6': 'ints',
            },
            name='return_many_ints',
            description=''
        )

        def return_four_ints() -> (list, list, list, list):
            return ([1, 2, 3], [4, 5, 6], [7, 8, 9], [4, 4])

        self.dp.methods.register_function(
            function=return_four_ints,
            inputs={},
            parameters={},
            outputs=[
                ('ints1', IntSequence1),
                ('ints2', IntSequence1),
                ('ints3', IntSequence1),
                ('ints4', IntSequence1),
            ],
            output_descriptions={
                'ints1': 'ints',
                'ints2': 'ints',
                'ints3': 'ints',
                'ints4': 'ints',
            },
            name='return_four_ints',
            description=''
        )

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_template_action_lumps_many_outputs(self):
        """
        ReplayPythonUsage._template_action should "lump" multiple outputs from
        one command into a single Results-like object when the total number of
        outputs from a single command > 5

        In these cases, our rendering should look like:
        `action_results = plugin_actions.action()...`
        instead of:
        `_, _, thing3, _, _, _ = plugin_actions.action()...`
        """
        ints = self.dp.actions['return_many_ints']()
        first_ints = ints[0]
        first_ints.save(os.path.join(self.tempdir, 'int-seq.qza'))
        fp = os.path.join(self.tempdir, 'int-seq.qza')
        out_fp = os.path.join(self.tempdir, 'action_collection.txt')
        replay_provenance(ReplayPythonUsage, fp, out_fp)

        exp = 'action_results = dummy_plugin_actions.return_many_ints'
        with open(out_fp) as fh:
            rendered = fh.read()
        self.assertRegex(rendered, exp)

    def test_template_action_does_not_lump_four_outputs(self):
        """
        ReplayPythonUsage._template_action should not "lump" multiple outputs
        one command into a single Results-like object when the total number of
        outputs from a single command <= 5, unless the total number of results
        is high (see above).

        In these cases, our rendering should look like:
        `_, _, thing3, _ = plugin_actions.action()...`
        instead of:
        `action_results = plugin_actions.action()...`
        """
        ints = self.dp.actions['return_four_ints']()
        first_ints = ints[0]
        first_ints.save(os.path.join(self.tempdir, 'int-seq.qza'))
        fp = os.path.join(self.tempdir, 'int-seq.qza')
        out_fp = os.path.join(self.tempdir, 'action_collection.txt')
        replay_provenance(ReplayPythonUsage, fp, out_fp)

        exp = 'ints1_0, _, _, _ = dummy_plugin_actions.return_four_ints'
        with open(out_fp) as fh:
            rendered = fh.read()
        self.assertRegex(rendered, exp)

    def test_template_action_lumps_three_variables(self):
        """
        ReplayPythonUsage._template_action should "lump" multiple outputs from
        one command into a single Results-like object when there are more than
        two usage variables (i.e. replay of 3+ results from a single command)

        In these cases, our rendering should look like:
        ```
        action_results = plugin_actions.action(...)
        thing1 = action_results.thinga
        etc.
        ```
        instead of:
        `thing1, _, thing3, _, thing5, _ = plugin_actions.action()...`
        """
        ints = self.dp.actions['return_four_ints']()
        os.mkdir(os.path.join(self.tempdir, 'three-ints-dir'))
        for i in range(3):
            out_path = os.path.join(self.tempdir, 'three-ints-dir',
                                    f'int-seq-{i}.qza')
            ints[i].save(out_path)

        fp = os.path.join(self.tempdir, 'three-ints-dir')
        out_fp = os.path.join(self.tempdir, 'action_collection.txt')
        replay_provenance(ReplayPythonUsage, fp, out_fp)

        exp = (
            'action_results = dummy_plugin_actions.return_four_ints',
            'ints1_0 = action_results.ints1',
            'ints2_0 = action_results.ints2',
            'ints3_0 = action_results.ints3'
        )
        with open(out_fp) as fh:
            rendered = fh.read()
        for pattern in exp:
            self.assertRegex(rendered, pattern)

    def test_template_action_does_not_lump_two_vars(self):
        """
        ReplayPythonUsage._template_action should not "lump" multiple outputs
        from one command into a single Results-like object when the total count
        of usage variables (i.e. replayed outputs) from a single command < 3,
        unless the total number of outputs is high (see above).

        In these cases, our rendering should look like:
        `thing1, _, thing3, _ = plugin_actions.action()...`
        instead of:
        `action_results = plugin_actions.action()...`
        """
        ints1, ints2, _, _ = self.dp.actions['return_four_ints']()
        os.mkdir(os.path.join(self.tempdir, 'two-ints-dir'))
        ints1.save(os.path.join(self.tempdir, 'two-ints-dir', 'int-seq-1.qza'))
        ints2.save(os.path.join(self.tempdir, 'two-ints-dir', 'int-seq-2.qza'))
        fp = os.path.join(self.tempdir, 'two-ints-dir')
        out_fp = os.path.join(self.tempdir, 'action_collection.txt')
        replay_provenance(ReplayPythonUsage, fp, out_fp)

        exp = 'ints1_0, ints2_0, _, _ = dummy_plugin_actions.return_four_ints'
        with open(out_fp) as fh:
            rendered = fh.read()
        self.assertRegex(rendered, exp)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import collections
import inspect
import copy
import itertools
import tempfile

import qiime2.sdk
import qiime2.core.type as qtype
from qiime2.core.archive.provenance import MetadataInfo
from .grammar import TypeExp, UnionExp
from .meta import TypeVarExp
from .collection import List, Set, Collection
from .primitive import infer_primitive_type
from .visualization import Visualization
from . import meta
from .util import (is_semantic_type, is_collection_type, is_primitive_type,
                   parse_primitive)
from ..util import ImmutableBase, md5sum, create_collection_name


class __NoValueMeta(type):
    def __repr__(self):
        return "NOVALUE"


# This sentinel is a class so that it retains the correct memory address when
# pickled
class _NOVALUE(metaclass=__NoValueMeta):
    pass


class ParameterSpec(ImmutableBase):
    NOVALUE = _NOVALUE

    def __init__(self, qiime_type=NOVALUE, view_type=NOVALUE, default=NOVALUE,
                 description=NOVALUE):
        self.qiime_type = qiime_type
        self.view_type = view_type
        self.default = default
        self.description = description

        self._freeze_()

    def has_qiime_type(self):
        return self.qiime_type is not self.NOVALUE

    def has_view_type(self):
        return self.view_type is not self.NOVALUE

    def has_default(self):
        return self.default is not self.NOVALUE

    def has_description(self):
        return self.description is not self.NOVALUE

    def duplicate(self, **kwargs):
        qiime_type = kwargs.pop('qiime_type', self.qiime_type)
        view_type = kwargs.pop('view_type', self.view_type)
        default = kwargs.pop('default', self.default)
        description = kwargs.pop('description', self.description)
        if kwargs:
            raise TypeError("Unknown arguments: %r" % kwargs)

        return ParameterSpec(qiime_type, view_type, default, description)

    def __repr__(self):
        return ("ParameterSpec(qiime_type=%r, view_type=%r, default=%r, "
                "description=%r)" % (self.qiime_type, self.view_type,
                                     self.default, self.description))

    def __eq__(self, other):
        return (self.qiime_type == other.qiime_type and
                self.view_type == other.view_type and
                self.default == other.default and
                self.description == other.description)

    def __ne__(self, other):
        return not (self == other)


class PipelineSignature:
    builtin_args = ('ctx',)

    def __init__(self, callable, inputs, parameters, outputs,
                 input_descriptions=None, parameter_descriptions=None,
                 output_descriptions=None):
        """

        Parameters
        ----------
        callable : callable
            Callable with view type annotations on parameters and return.
        inputs : dict
            Parameter name to semantic type.
        parameters : dict
            Parameter name to primitive type.
        outputs : dict or list of tuples
            Each pair/tuple contains the name of the output (str) and its QIIME
            type.
        input_descriptions : dict, optional
            Input name to description string.
        parameter_descriptions : dict, optional
            Parameter name to description string.
        output_descriptions : dict, optional
            Output name to description string.

        """
        # update type of outputs if needed
        if type(outputs) is list:
            outputs = dict(outputs)
        elif type(outputs) is set:
            raise ValueError("Plugin registration for %r cannot use a set()"
                             " to define the outputs, as the order is random."
                             % callable.__name__)

        inputs, parameters, outputs, signature_order = \
            self._parse_signature(callable, inputs, parameters, outputs,
                                  input_descriptions, parameter_descriptions,
                                  output_descriptions)

        self._assert_valid_inputs(inputs)
        self._assert_valid_parameters(parameters)
        self._assert_valid_outputs(outputs)
        self._assert_valid_views(inputs, parameters, outputs)
        self.inputs = inputs
        self.parameters = parameters
        self.outputs = outputs
        self.signature_order = signature_order

    def _parse_signature(self, callable, inputs, parameters, outputs,
                         input_descriptions=None, parameter_descriptions=None,
                         output_descriptions=None):
        #  Initialize dictionaries if non-existant.
        if input_descriptions is None:
            input_descriptions = {}
        if parameter_descriptions is None:
            parameter_descriptions = {}
        if output_descriptions is None:
            output_descriptions = {}

        # Copy so we can "exhaust" the collections and check for missing params
        inputs = copy.copy(inputs)
        parameters = copy.copy(parameters)
        input_descriptions = copy.copy(input_descriptions)
        parameter_descriptions = copy.copy(parameter_descriptions)
        output_descriptions = copy.copy(output_descriptions)
        builtin_args = list(self.builtin_args)

        annotated_inputs = collections.OrderedDict()
        annotated_parameters = collections.OrderedDict()
        annotated_outputs = collections.OrderedDict()
        signature_order = collections.OrderedDict()

        for name, parameter in inspect.signature(callable).parameters.items():
            if (parameter.kind == parameter.VAR_POSITIONAL or
                    parameter.kind == parameter.VAR_KEYWORD):
                raise TypeError("Variadic definitions are unsupported: %r" %
                                name)

            if builtin_args:
                if builtin_args[0] != name:
                    raise TypeError("Missing builtin argument %r, got %r" %
                                    (builtin_args[0], name))
                builtin_args = builtin_args[1:]
                continue

            view_type = ParameterSpec.NOVALUE
            if parameter.annotation is not parameter.empty:
                view_type = parameter.annotation
            default = ParameterSpec.NOVALUE
            if parameter.default is not parameter.empty:
                default = parameter.default

            if name in inputs:
                description = input_descriptions.pop(name,
                                                     ParameterSpec.NOVALUE)
                param_spec = ParameterSpec(
                    qiime_type=inputs.pop(name), view_type=view_type,
                    default=default, description=description)
                annotated_inputs[name] = param_spec
                signature_order[name] = param_spec
            elif name in parameters:
                description = parameter_descriptions.pop(name,
                                                         ParameterSpec.NOVALUE)
                param_spec = ParameterSpec(
                    qiime_type=parameters.pop(name), view_type=view_type,
                    default=default, description=description)
                annotated_parameters[name] = param_spec
                signature_order[name] = param_spec
            elif name not in self.builtin_args:
                raise TypeError("Parameter in callable without QIIME type:"
                                " %r" % name)
        # we should have popped both of these empty by this point
        if inputs or parameters:
            raise TypeError("Callable does not have parameter(s): %r"
                            % (list(inputs) + list(parameters)))

        if 'return' in callable.__annotations__:
            output_views = qiime2.core.util.tuplize(
                callable.__annotations__['return'])

            if len(output_views) != len(outputs):
                raise TypeError("Number of registered outputs (%r) does not"
                                " match annotation (%r)" %
                                (len(outputs), len(output_views)))

            for (name, qiime_type), view_type in zip(outputs.items(),
                                                     output_views):
                description = output_descriptions.pop(name,
                                                      ParameterSpec.NOVALUE)
                annotated_outputs[name] = ParameterSpec(
                    qiime_type=qiime_type, view_type=view_type,
                    description=description)
        else:
            for name, qiime_type in outputs.items():
                description = output_descriptions.pop(name,
                                                      ParameterSpec.NOVALUE)
                annotated_outputs[name] = ParameterSpec(
                    qiime_type=qiime_type, description=description)

        # we should have popped the descriptions empty by this point
        if input_descriptions or parameter_descriptions or output_descriptions:
            raise TypeError(
                "Callable does not have parameter(s)/output(s) found in "
                "descriptions: %r" % [*input_descriptions,
                                      *parameter_descriptions,
                                      *output_descriptions])

        return (annotated_inputs, annotated_parameters, annotated_outputs,
                signature_order)

    def collate_inputs(self, *args, **kwargs):
        # Collate positional inputs
        collated_inputs = {name: value for name, value in
                           zip(self.signature_order, args)}
        collated_inputs.update(kwargs)
        return collated_inputs

    def _assert_valid_inputs(self, inputs):
        for input_name, spec in inputs.items():
            if not is_semantic_type(spec.qiime_type):
                raise TypeError(
                    "Input %r must be a semantic QIIME type, not %r"
                    % (input_name, spec.qiime_type))

            if not isinstance(spec.qiime_type, (TypeExp, UnionExp)):
                raise TypeError(
                    "Input %r must be a complete semantic type expression, "
                    "not %r" % (input_name, spec.qiime_type))

            if spec.has_default() and spec.default is not None:
                raise ValueError(
                    "Input %r has a default value of %r. Only a default "
                    "value of `None` is supported for inputs."
                    % (input_name, spec.default))

            for var_selector in meta.select_variables(spec.qiime_type):
                var = var_selector(spec.qiime_type)
                if not var.input:
                    raise TypeError("An output variable has been associated"
                                    " with an input type: %r"
                                    % spec.qiime_type)

    def _assert_valid_parameters(self, parameters):
        for param_name, spec in parameters.items():
            if not is_primitive_type(spec.qiime_type):
                raise TypeError(
                    "Parameter %r must be a primitive QIIME type, not %r"
                    % (param_name, spec.qiime_type))

            if not isinstance(spec.qiime_type, (TypeExp, UnionExp)):
                raise TypeError(
                    "Parameter %r must be a complete primitive type "
                    "expression, not %r" % (param_name, spec.qiime_type))

            if (spec.has_default() and
                    spec.default is not None and
                    spec.default not in spec.qiime_type):
                raise TypeError("Default value for parameter %r is not of "
                                "semantic QIIME type %r or `None`."
                                % (param_name, spec.qiime_type))

            for var_selector in meta.select_variables(spec.qiime_type):
                var = var_selector(spec.qiime_type)
                if not var.input:
                    raise TypeError("An output variable has been associated"
                                    " with an input type: %r"
                                    % spec.qiime_type)

    def _assert_valid_outputs(self, outputs):
        if len(outputs) == 0:
            raise TypeError("%s requires at least one output"
                            % self.__class__.__name__)

        for output_name, spec in outputs.items():
            if not (is_semantic_type(spec.qiime_type) or
                    spec.qiime_type == Visualization or
                    spec.qiime_type == Collection[Visualization]):
                raise TypeError(
                    "Output %r must be a semantic QIIME type or "
                    "Visualization, not %r"
                    % (output_name, spec.qiime_type))

            if not isinstance(spec.qiime_type, (TypeVarExp, TypeExp)):
                raise TypeError(
                    "Output %r must be a complete type expression, not %r"
                    % (output_name, spec.qiime_type))

            for var_selector in meta.select_variables(spec.qiime_type):
                var = var_selector(spec.qiime_type)
                if not var.output:
                    raise TypeError("An input variable has been associated"
                                    " with an input type: %r")

    def _assert_valid_views(self, inputs, parameters, outputs):
        for name, spec in itertools.chain(inputs.items(),
                                          parameters.items(),
                                          outputs.items()):
            if spec.has_view_type():
                raise TypeError(
                    " Pipelines do not support function annotations (found one"
                    " for parameter: %r)." % name)

    def coerce_user_input(self, **user_input):
        """ Coerce user inputs to be appropriate for callable
        """
        callable_args = {}

        for name, spec in self.signature_order.items():
            # Some arguments may be optional and won't be present here. Whether
            # they passed all mandatory arguments or not is validated elsewhere
            if name in user_input:
                arg = user_input[name]
                if name in self.inputs:
                    callable_args[name] = self._coerce_given_input(arg, spec)
                else:
                    callable_args[name] = \
                        self._coerce_given_parameter(arg, spec)

        return callable_args

    def _coerce_given_input(self, _input, spec):
        """ Coerce input to be appropriate for callable
        """
        _, qiime_name = self._get_qiime_type_and_name(spec)

        # Transform collection from list to dict and vice versa if needed
        if qiime_name == 'Collection' and isinstance(_input, list):
            _input = self._list_to_dict(_input)
        elif qiime_name == 'List' and \
                (isinstance(_input, dict) or
                 isinstance(_input, qiime2.sdk.ResultCollection)):
            _input = self._dict_to_list(_input)

        if isinstance(_input, dict):
            _input = qiime2.sdk.ResultCollection(_input)

        return _input

    def _coerce_given_parameter(self, param, spec):
        """ Coerce parameter to be appropriate for callable
        """
        view_type = spec.view_type
        qiime_type = spec.qiime_type

        if view_type == dict and isinstance(param, list):
            param = self._list_to_dict(param)
        elif view_type == list and isinstance(param, dict):
            param = self._dict_to_list(param)

        if qiime_type is qtype.Threads and param == 'auto':
            param = 0

        return param

    def transform_and_add_callable_args_to_prov(self, provenance,
                                                **callable_args):
        """ Transform inputs to views and add all callable arguments to
            provenance. Needs to be done together so we can add transformation
            records to provenance and because we want transformers to run
            outside the DFK in parsl
        """
        for name, spec in self.signature_order.items():
            arg = callable_args[name]

            if name in self.inputs:
                callable_args[name] = \
                    self._transform_and_add_input_to_prov(
                        provenance, name, spec, arg)
            else:
                provenance.add_parameter(name, spec.qiime_type, arg)

        return callable_args

    def _transform_and_add_input_to_prov(self, provenance, name, spec, _input):
        """ Transform the input and add both the input and the transformation
            record to provenance
        """
        transformed_input = None

        # Add input to provenance after creating the correct collection
        # type
        provenance.add_input(name, _input)
        qiime_type, _ = self._get_qiime_type_and_name(spec)

        # Transform artifacts to view types as necessary
        if _input is None:
            transformed_input = None
        elif spec.has_view_type():
            recorder = provenance.transformation_recorder(name)
            # Transform all members of collection into view type
            if qtype.is_collection_type(qiime_type):
                if isinstance(_input, qiime2.sdk.result.ResultCollection):
                    transformed_input = qiime2.sdk.result.ResultCollection(
                        {k: v._view(spec.view_type,
                                    recorder) for k, v in _input.items()})
                else:
                    transformed_input = [
                        i._view(spec.view_type, recorder) for i in _input]
            else:
                transformed_input = _input._view(spec.view_type, recorder)
        else:
            transformed_input = _input

        return transformed_input

    def _get_qiime_type_and_name(self, spec):
        """ Get concrete qiime type and name from nested spec
        """
        qiime_type = spec.qiime_type
        qiime_name = spec.qiime_type.name

        # I don't think this will necessarily work if we nest collection
        # types in the future
        if qiime_name == '':
            # If we have an outer union as our semantic type, the name will
            # be the empty string, and the type will be the entire union
            # expression. In order to get a meaningful name and a type
            # that tells us if we have a collection, we unpack the union
            # and grab that info from the first element. All subsequent
            # elements will share this same basic information because we
            # do not allow
            # List[TypeA] | Collection[TypeA]
            qiime_type = next(iter(spec.qiime_type))
            qiime_name = qiime_type.name

        return qiime_type, qiime_name

    def coerce_given_outputs(self, output_views, output_types, scope,
                             provenance):
        """ Coerce the outputs produced by the method into the desired types if
            possible. Primarily useful to create collections of outputs
        """
        outputs = []

        for output_view, (name, spec) in zip(output_views,
                                             output_types.items()):
            if spec.qiime_type.name == 'Collection':
                output = qiime2.sdk.ResultCollection()
                size = len(output_view)

                if isinstance(output_view, qiime2.sdk.ResultCollection) or \
                        isinstance(output_view, dict):
                    keys = list(output_view.keys())
                    values = list(output_view.values())
                else:
                    keys = None
                    values = output_view

                for idx, view in enumerate(values):
                    if keys is not None:
                        key = str(keys[idx])
                    else:
                        key = str(idx)

                    collection_name = create_collection_name(
                        name=name, key=key, idx=idx, size=size)
                    output[key] = self._create_output_artifact(
                        provenance, collection_name, scope, spec, view)
            elif type(output_view) is not spec.view_type:
                raise TypeError(
                    "Expected output view type %r, received %r" %
                    (spec.view_type.__name__, type(output_view).__name__))
            else:
                output = self._create_output_artifact(
                    provenance, name, scope, spec, output_view)

            outputs.append(output)

        return outputs

    def _create_output_artifact(self, provenance, name, scope, spec, view):
        """ Create an output artifact from a view and add it to provenance
        """
        prov = provenance.fork(name)
        qiime_type = spec.qiime_type

        # If we have a collection we need to get a concrete qiime_type to
        # instantiate each artifact as.
        #
        # For instance, we cannot instantiate a Collection[SingleInt] from an
        # integer. We want to instantiate a SingleInt that will be put into a
        # ResultCollection outside of this method.
        if is_collection_type(qiime_type):
            qiime_type = qiime_type.fields[0]

        scope.add_reference(prov)

        artifact = qiime2.sdk.Artifact._from_view(
            qiime_type, view, spec.view_type, prov)
        artifact = scope.add_parent_reference(artifact)

        return artifact

    def decode_parameters(self, **kwargs):
        params = {}
        for key, spec in self.parameters.items():
            if (spec.has_default() and
                    spec.default is None and
                    kwargs[key] is None):
                params[key] = None
            else:
                params[key] = parse_primitive(spec.qiime_type, kwargs[key])
        return params

    def _dict_to_list(self, _input):
        """ Turn dict to list
        """
        return list(_input.values())

    def _list_to_dict(self, _input):
        """ Turn list to dict
        """
        return {str(idx): v for idx, v in enumerate(_input)}

    def check_types(self, **kwargs):
        for name, spec in self.signature_order.items():
            parameter = kwargs[name]
            # A type mismatch is unacceptable unless the value is None
            # and this parameter's default value is None.
            if ((parameter not in spec.qiime_type) and
                    not (spec.has_default() and spec.default is None
                         and parameter is None)):

                if isinstance(parameter, qiime2.sdk.Visualization):
                    raise TypeError(
                        "Parameter %r received a Visualization as an "
                        "argument. Visualizations may not be used as inputs."
                        % name)

                elif isinstance(parameter, qiime2.sdk.Artifact):
                    raise TypeError(
                        "Parameter %r requires an argument of type %r. An "
                        "argument of type %r was passed." % (
                            name, spec.qiime_type, parameter.type))

                elif isinstance(parameter, qiime2.Metadata):
                    raise TypeError(
                        "Parameter %r received Metadata as an "
                        "argument, which is incompatible with parameter "
                        "type: %r" % (name, spec.qiime_type))

                else:  # handle primitive types
                    raise TypeError(
                        "Parameter %r received %r as an argument, which is "
                        "incompatible with parameter type: %r"
                        % (name, parameter, spec.qiime_type))

    def solve_output(self, **kwargs):
        solved_outputs = None
        for _, spec in itertools.chain(self.inputs.items(),
                                       self.parameters.items(),
                                       self.outputs.items()):
            if list(meta.select_variables(spec.qiime_type)):
                break  # a variable exists, do the hard work
        else:
            # no variables
            solved_outputs = self.outputs

        if solved_outputs is None:
            inputs = {**{k: s.qiime_type for k, s in self.inputs.items()},
                      **{k: s.qiime_type for k, s in self.parameters.items()}}
            outputs = {k: s.qiime_type for k, s in self.outputs.items()}
            input_types = {
                k: self._infer_type(k, v) for k, v in kwargs.items()}

            solved = meta.match(input_types, inputs, outputs)
            solved_outputs = collections.OrderedDict(
                (k, s.duplicate(qiime_type=solved[k]))
                for k, s in self.outputs.items())

        for output_name, spec in solved_outputs.items():
            if not spec.qiime_type.is_concrete():
                raise TypeError(
                    "Solved output %r must be a concrete type, not %r" %
                    (output_name, spec.qiime_type))

        return solved_outputs

    def _infer_type(self, key, value):
        if value is None:
            if key in self.inputs:
                return self.inputs[key].qiime_type
            elif key in self.parameters:
                return self.parameters[key].qiime_type
            # Shouldn't happen:
            raise ValueError("Parameter passed not consistent with signature.")
        if type(value) is list:
            inner = UnionExp((self._infer_type(key, v) for v in value))
            return List[inner.normalize()]
        if type(value) is set:
            inner = UnionExp((self._infer_type(key, v) for v in value))
            return Set[inner.normalize()]
        if type(value) is dict or \
                isinstance(value, qiime2.sdk.ResultCollection):
            inner = UnionExp(
                (self._infer_type(key, v) for v in value.values()))
            return Collection[inner.normalize()]
        if isinstance(
                value, (qiime2.sdk.Artifact, qiime2.sdk.proxy.ProxyArtifact)):
            return value.type
        else:
            return infer_primitive_type(value)

    def __repr__(self):
        lines = []
        for group in 'inputs', 'parameters', 'outputs':
            lookup = getattr(self, group)
            lines.append('%s:' % group)
            for name, spec in lookup.items():
                lines.append('    %s: %r' % (name, spec))
        return '\n'.join(lines)

    def __eq__(self, other):
        return (type(self) is type(other) and
                self.inputs == other.inputs and
                self.parameters == other.parameters and
                self.outputs == other.outputs and
                self.signature_order == other.signature_order)

    def __ne__(self, other):
        return not (self == other)


class MethodSignature(PipelineSignature):
    builtin_args = ()

    def _assert_valid_outputs(self, outputs):
        super()._assert_valid_outputs(outputs)
        # Assert all output types are semantic types. The parent class is less
        # strict in its output type requirements.
        for output_name, spec in outputs.items():
            if not is_semantic_type(spec.qiime_type):
                raise TypeError(
                    "Output %r must be a semantic QIIME type, not %r" %
                    (output_name, spec.qiime_type))

    def _assert_valid_views(self, inputs, parameters, outputs):
        for name, spec in itertools.chain(inputs.items(),
                                          parameters.items(),
                                          outputs.items()):
            if not spec.has_view_type():
                raise TypeError("Method is missing a function annotation for"
                                " parameter: %r" % name)


class VisualizerSignature(PipelineSignature):
    builtin_args = ('output_dir',)

    def __init__(self, callable, inputs, parameters, input_descriptions=None,
                 parameter_descriptions=None):
        outputs = {'visualization': Visualization}
        output_descriptions = None
        super().__init__(callable, inputs, parameters, outputs,
                         input_descriptions, parameter_descriptions,
                         output_descriptions)

    def _assert_valid_outputs(self, outputs):
        super()._assert_valid_outputs(outputs)
        output = outputs['visualization']
        if output.has_view_type() and output.view_type is not None:
            raise TypeError(
                "Visualizer callable cannot return anything. Its return "
                "annotation must be `None`, not %r. Write output to "
                "`output_dir`." % output.view_type)

    def _assert_valid_views(self, inputs, parameters, outputs):
        for name, spec in itertools.chain(inputs.items(), parameters.items()):
            if not spec.has_view_type():
                raise TypeError("Visualizer is missing a function annotation"
                                " for parameter: %r" % name)


IndexedCollectionElement = collections.namedtuple(
    'IndexedCollectionElement', ['item_name', 'idx', 'total'])


class HashableInvocation():
    def __init__(self, plugin_action, arguments):
        self.plugin_action = plugin_action

        unified_arguments = self._unify_dicts(arguments)
        self.arguments = self._make_hashable(unified_arguments)

    def __eq__(self, other):
        return (self.plugin_action == other.plugin_action) \
              and (self.arguments == other.arguments)

    def __hash__(self):
        return hash((self.plugin_action, self.arguments))

    def __repr__(self):
        return (f'\nPLUGIN_ACTION: {self.plugin_action}\nARGUMENTS:'
                f' {self.arguments}\n')

    def _unify_dicts(self, arguments):
        """Check if action.yaml gave us any lists of single element dicts to
        unify
        """
        for idx, argument in enumerate(arguments):
            name, value = list(argument.items())[0]
            if isinstance(value, list) and \
                    all(isinstance(x, dict) for x in value):
                arguments[idx] = {name: self._unify_dict(value)}

        return arguments

    def _unify_dict(self, collection):
        """If we do have a list of single element dicts, turn it into one dict
        """
        unified_dict = {}

        for elem in collection:
            for k, v in elem.items():
                unified_dict[k] = v

        return unified_dict

    def _make_hashable(self, collection):
        """Take an arbitrarily nested collection and turn it into a hashable
        arbitrarily nested tuple. Turns Artifacts into their uuid and Metadata
        into their md5sum
        """
        from qiime2 import Artifact
        from qiime2.sdk import ResultCollection
        from qiime2.metadata.metadata import _MetadataBase

        new_collection = []

        if isinstance(collection, dict) or \
                isinstance(collection, ResultCollection):
            for k, v in collection.items():
                new_collection.append((k, self._make_hashable(v)))
        elif isinstance(collection, list):
            for elem in collection:
                new_collection.append(self._make_hashable(elem))
        elif isinstance(collection, Artifact):
            return str(collection.uuid)
        elif isinstance(collection, _MetadataBase):
            with tempfile.NamedTemporaryFile('w') as fh:
                fp = fh.name
                collection.save(fp)
                collection = md5sum(fp)
                return collection
        elif isinstance(collection, MetadataInfo):
            return collection.md5sum_hash
        else:
            return collection

        return tuple(new_collection)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import io
import os
import pathlib
import random
import shutil
import tempfile
import unittest
import zipfile
from contextlib import redirect_stdout

import networkx as nx
from networkx import DiGraph
import pytest

from .._checksum_validator import ValidationCode
from ..parse import (
    ProvDAG, DirectoryParser, EmptyParser, ProvDAGParser, select_parser,
    parse_provenance, UnparseableDataError
)
from ..archive_parser import (
    ParserV0, ParserV1, ParserV2, ParserV3, ParserV4, ParserV5, ParserV6,
    Config, ProvNode, ParserResults, ArchiveParser,
)

from .testing_utilities import (
    is_root_provnode_data, generate_archive_with_file_removed, DummyArtifacts
)

from qiime2 import Artifact
from qiime2.core.archive.archiver import ChecksumDiff


class ProvDAGTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def save_artifact_to_dir(self, artifact, directory):
        dir_fp = os.path.join(self.tempdir, directory)
        try:
            os.mkdir(dir_fp)
        except FileExistsError:
            pass

        fp = os.path.join(dir_fp, f'{artifact.name}.qza')
        artifact.artifact.save(fp)
        return dir_fp

    def test_number_of_nodes(self):
        num_single_int_nodes = len(self.das.single_int.dag.nodes)
        self.assertEqual(num_single_int_nodes, 1)

        num_int_seq1_nodes = len(self.das.int_seq1.dag.nodes)
        self.assertEqual(num_int_seq1_nodes, 1)

        num_concated_ints_nodes = len(self.das.concated_ints.dag.nodes)
        self.assertEqual(num_concated_ints_nodes, 3)

    def test_number_of_nodes_pipeline(self):
        # input int-seq, input mapping, pipeline output viz, aliased viz,
        # split int-seq
        num_pipeline_viz_nodes = len(self.das.pipeline_viz.dag.nodes)
        self.assertEqual(num_pipeline_viz_nodes, 5)

    def test_number_of_terminal_nodes(self):
        num_int_seq1_term_nodes = len(self.das.int_seq1.dag.terminal_nodes)
        self.assertEqual(num_int_seq1_term_nodes, 1)

        num_concated_ints_term_nodes = \
            len(self.das.concated_ints.dag.terminal_nodes)
        self.assertEqual(num_concated_ints_term_nodes, 1)

        self.save_artifact_to_dir(self.das.int_seq1, 'two-ints')
        dir_fp = self.save_artifact_to_dir(self.das.int_seq2, 'two-ints')
        dag = ProvDAG(dir_fp)
        self.assertEqual(len(dag.terminal_nodes), 2)

        self.save_artifact_to_dir(self.das.int_seq1, 'int-and-concated-ints')
        dir_fp = self.save_artifact_to_dir(self.das.concated_ints,
                                           'int-and-concated-ints')
        dag = ProvDAG(dir_fp)
        self.assertEqual(len(dag.terminal_nodes), 1)

    def test_number_of_terminal_nodes_pipeline(self):
        num_pipeline_viz_term_nodes = \
            len(self.das.pipeline_viz.dag.terminal_nodes)
        self.assertEqual(num_pipeline_viz_term_nodes, 1)

    def test_root_node_is_archive_root(self):
        with zipfile.ZipFile(self.das.concated_ints.filepath) as zf:
            all_filenames = zf.namelist()
            root_filenames = filter(is_root_provnode_data, all_filenames)
            root_filepaths = [pathlib.Path(fp) for fp in root_filenames]
            exp_node = ProvNode(Config(), zf, root_filepaths)
            act_terminal_node, *_ = self.das.concated_ints.dag.terminal_nodes
            self.assertEqual(exp_node, act_terminal_node)

    def test_number_of_actions(self):
        self.assertEqual(self.das.int_seq1.dag.dag.number_of_edges(), 0)

        self.assertEqual(self.das.concated_ints.dag.dag.number_of_edges(), 2)

    def test_number_of_actions_pipeline(self):
        # (1) one edge from input intseq to pipeline output viz
        # (2) one edge from input mapping to pipeline output viz
        # (3) one edge from input intseq to left split int
        # (4) one edge from left split int to true output viz, of which
        # pipeline output viz is an alias
        self.assertEqual(self.das.pipeline_viz.dag.dag.number_of_edges(), 4)

    def test_nonexistent_fp(self):
        fp = os.path.join(self.tempdir, 'does-not-exist.qza')
        err_msg = 'FileNotFoundError'
        with self.assertRaisesRegex(UnparseableDataError, err_msg):
            ProvDAG(fp)

    def test_insufficient_permissions(self):
        fp = os.path.join(self.tempdir, 'int-seq-1-permissions-copy.qza')
        self.das.int_seq1.artifact.save(fp)

        os.chmod(fp, 0o000)
        err_msg = 'PermissionError.*Permission denied'
        with self.assertRaisesRegex(UnparseableDataError, err_msg):
            ProvDAG(fp)

        os.chmod(fp, 0o123)
        err_msg = 'PermissionError.*Permission denied'
        with self.assertRaisesRegex(UnparseableDataError, err_msg):
            ProvDAG(fp)

    def test_not_a_zip_file(self):
        fp = os.path.join(self.tempdir, 'not-a-zip.txt')
        with open(fp, 'w') as fh:
            fh.write("This is just a text file.")

        err_msg = 'zipfile.BadZipFile.*File is not a zip file'
        with self.assertRaisesRegex(UnparseableDataError, err_msg):
            ProvDAG(fp)

    def test_has_digraph(self):
        self.assertIsInstance(self.das.int_seq1.dag.dag, DiGraph)

        self.assertIsInstance(self.das.concated_ints.dag.dag, DiGraph)

    def test_dag_attributes(self):
        dag = self.das.int_seq1.dag
        terminal_node, *_ = dag.terminal_nodes
        self.assertIsInstance(terminal_node, ProvNode)

        self.assertEqual(dag.provenance_is_valid, ValidationCode.VALID)

        empty_checksum_diff = ChecksumDiff(added={}, removed={}, changed={})
        self.assertEqual(dag.checksum_diff, empty_checksum_diff)

    def test_node_action_names(self):
        int_seq1_node, *_ = self.das.int_seq1.dag.terminal_nodes
        self.assertEqual(int_seq1_node.action.action_name, 'import')

        concated_ints_node, *_ = self.das.concated_ints.dag.terminal_nodes
        self.assertEqual(concated_ints_node.action.action_name,
                         'concatenate_ints')

    def test_node_action_names_pipeline(self):
        pipeline_viz_node, *_ = self.das.pipeline_viz.dag.terminal_nodes
        self.assertEqual(pipeline_viz_node.action.action_name,
                         'typical_pipeline')

    def test_has_correct_edges(self):
        edges = self.das.concated_ints.dag.dag.edges

        self.assertIn(
            (self.das.int_seq1.uuid, self.das.concated_ints.uuid), edges
        )
        self.assertIn(
            (self.das.int_seq2.uuid, self.das.concated_ints.uuid), edges
        )
        self.assertNotIn(
            (self.das.int_seq1.uuid, self.das.int_seq2.uuid), edges
        )
        self.assertNotIn(
            (self.das.concated_ints.uuid, self.das.int_seq1.uuid), edges
        )

    def test_dag_repr(self):
        exp_repr = f'ProvDAG.*Artifacts.*{self.das.int_seq1.uuid}'
        self.assertRegex(repr(self.das.int_seq1.dag), exp_repr)

    def test_node_repr(self):
        int_seq1_node, *_ = self.das.int_seq1.dag.terminal_nodes
        uuid = int_seq1_node._uuid
        type_ = int_seq1_node.type
        format_ = int_seq1_node.format

        exp_repr = f'(?s)UUID.*{uuid}.*Type:.*{type_}.*Data Format:.*{format_}'
        self.assertRegex(repr(int_seq1_node), exp_repr)

    def test_dag_eq(self):
        dag = self.das.int_seq1.dag
        self.assertEqual(dag, dag)

        fp = self.das.int_seq1.filepath
        self.assertEqual(ProvDAG(fp), ProvDAG(fp))

        # because they are isomorphic
        self.assertEqual(self.das.int_seq1.dag, self.das.int_seq2.dag)

    def test_dag_not_eq(self):
        self.assertNotEqual(self.das.int_seq1.dag, self.das.concated_ints.dag)

    def test_captures_full_history(self):
        concat_ints = self.das.dp.actions['concatenate_ints']

        next_concated_ints = self.das.concated_ints.artifact
        iterations = random.randint(1, 10)
        for _ in range(iterations):
            next_concated_ints, = concat_ints(next_concated_ints,
                                              next_concated_ints,
                                              self.das.int_seq2.artifact,
                                              4, 6)

        fp = os.path.join(self.tempdir, 'very-concated-ints.qza')
        next_concated_ints.save(fp)
        dag = ProvDAG(fp)
        # iterations + o.g. concated_ints + o.g. int_seq + o.g. int_seq2
        self.assertEqual(len(dag), iterations + 3)

    def test_get_outer_provenance_nodes(self):
        fp = os.path.join(self.tempdir, 'disconnected-provenances')
        os.mkdir(fp)
        self.das.concated_ints.artifact.save(
            os.path.join(fp, 'concated-ints.qza'))
        unattached_int_seq = Artifact.import_data('IntSequence1', [8, 8])
        unattached_int_seq.save(os.path.join(fp, 'unattached-int-seq.qza'))
        dag = ProvDAG(fp)

        actual = dag.get_outer_provenance_nodes(self.das.concated_ints.uuid)
        exp = {
            self.das.concated_ints.uuid,
            self.das.int_seq1.uuid,
            self.das.int_seq2.uuid
        }
        self.assertEqual(actual, exp)

    def test_get_outer_provenance_nodes_pipeline(self):
        dag = self.das.pipeline_viz.dag
        actual = dag.get_outer_provenance_nodes(self.das.pipeline_viz.uuid)
        exp = {
            self.das.pipeline_viz.uuid,
            self.das.int_seq1.uuid,
            self.das.mapping1.uuid
        }
        self.assertEqual(actual, exp)

    def test_collapsed_view(self):
        view = self.das.concated_ints.dag.collapsed_view
        self.assertIsInstance(view, DiGraph)
        self.assertEqual(len(view), 3)

        exp_nodes = [
            self.das.concated_ints.uuid,
            self.das.int_seq1.uuid,
            self.das.int_seq2.uuid
        ]
        for exp_node in exp_nodes:
            self.assertIn(exp_node, view.nodes)

    def test_collapsed_view_pipeline(self):
        view = self.das.pipeline_viz.dag.collapsed_view
        self.assertIsInstance(view, DiGraph)
        self.assertEqual(len(view), 3)

        exp_nodes = [
            self.das.pipeline_viz.uuid,
            self.das.int_seq1.uuid,
            self.das.mapping1.uuid
        ]
        for exp_node in exp_nodes:
            self.assertIn(exp_node, view.nodes)

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_invalid_provenance(self):
        '''
        Mangle an intact v5 Archive so that its checksums.md5 is invalid,
        and then build a ProvDAG with it to confirm the ProvDAG constructor
        handles broken checksums appropriately
        '''
        uuid = self.das.int_seq1.uuid
        with generate_archive_with_file_removed(
            self.das.int_seq1.filepath,
            uuid,
            os.path.join('data', 'ints.txt')
        ) as altered_archive:
            new_fp = os.path.join(uuid, 'data', 'tamper.txt')
            overwrite_fp = os.path.join(uuid, 'provenance', 'citations.bib')
            with zipfile.ZipFile(altered_archive, 'a') as zf:
                zf.writestr(new_fp, 'added file')

                with zf.open(overwrite_fp, 'w') as fh:
                    fh.write(b'999\n')

            expected = (
                '(?s)'
                f'Checksums are invalid for Archive {uuid}.*'
                'Archive may be corrupt.*'
                'Files added.*tamper.txt.*'
                'Files removed.*ints.txt.*'
                'Files changed.*provenance.*citations.bib.*'
            )

            with self.assertWarnsRegex(UserWarning, expected):
                dag = ProvDAG(altered_archive)

            self.assertEqual(dag.provenance_is_valid, ValidationCode.INVALID)

            diff = dag.checksum_diff
            self.assertEqual(list(diff.removed.keys()), ['data/ints.txt'])
            self.assertEqual(list(diff.added.keys()), ['data/tamper.txt'])
            self.assertEqual(list(diff.changed.keys()),
                             ['provenance/citations.bib'])

    def test_missing_checksums_md5(self):
        uuid = self.das.single_int.uuid
        with generate_archive_with_file_removed(
            self.das.single_int.filepath,
            uuid,
            'checksums.md5'
        ) as altered_archive:
            expected = (
                'The checksums.md5 file is missing from the archive.*'
                'Archive may be corrupt'
            )
            with self.assertWarnsRegex(UserWarning, expected):
                dag = ProvDAG(altered_archive)

            self.assertEqual(dag.provenance_is_valid, ValidationCode.INVALID)

            diff = dag.checksum_diff
            self.assertEqual(diff, None)

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_error_if_missing_node_files(self):
        path_prefix = os.path.join('provenance', 'artifacts')
        root_uuid = self.das.concated_ints.uuid
        for removed_file in [
            'metadata.yaml',
            'citations.bib',
            'VERSION',
            'action/action.yaml'
        ]:
            for uuid in [self.das.int_seq1.uuid, self.das.int_seq2.uuid]:
                with generate_archive_with_file_removed(
                    self.das.concated_ints.filepath,
                    root_uuid,
                    os.path.join(path_prefix, uuid, removed_file)
                ) as altered_archive:
                    if removed_file == 'action/action.yaml':
                        file = 'action.yaml'
                    else:
                        file = removed_file

                    expected = (
                        f'(?s)Malformed.*{file}.*{uuid}.*corrupt.*'
                    )
                    with self.assertRaisesRegex(ValueError, expected):
                        ProvDAG(altered_archive)

    def test_v0_archive(self):
        dag = self.das.table_v0.dag
        uuid = self.das.table_v0.uuid

        self.assertEqual(
            dag.provenance_is_valid, ValidationCode.PREDATES_CHECKSUMS
        )
        self.assertEqual(dag.node_has_provenance(uuid), False)

    def test_v1_archive(self):
        dag = self.das.concated_ints_v1.dag
        uuid = self.das.concated_ints_v1.uuid

        self.assertEqual(
            dag.provenance_is_valid, ValidationCode.PREDATES_CHECKSUMS
        )
        self.assertEqual(dag.node_has_provenance(uuid), False)

    def test_v2_archive(self):
        dag = self.das.concated_ints_v2.dag
        uuid = self.das.concated_ints_v2.uuid

        self.assertEqual(
            dag.provenance_is_valid, ValidationCode.PREDATES_CHECKSUMS
        )
        self.assertEqual(dag.node_has_provenance(uuid), True)

    def test_v4_archive(self):
        dag = self.das.concated_ints_v4.dag
        uuid = self.das.concated_ints_v4.uuid

        self.assertEqual(
            dag.provenance_is_valid, ValidationCode.PREDATES_CHECKSUMS
        )
        self.assertEqual(dag.node_has_provenance(uuid), True)

        with zipfile.ZipFile(self.das.concated_ints_v4.filepath) as zf:
            citations_path = os.path.join(uuid, 'provenance', 'citations.bib')
            self.assertIn(citations_path, zf.namelist())

    def test_v5_archive(self):
        dag = self.das.concated_ints_v5.dag
        uuid = self.das.concated_ints_v5.uuid

        self.assertEqual(dag.provenance_is_valid, ValidationCode.VALID)
        self.assertEqual(dag.node_has_provenance(uuid), True)

    def test_artifact_passed_as_metadata_archive(self):
        dag = self.das.mapping1.dag
        uuid = self.das.mapping1.uuid
        self.assertEqual(dag.node_has_provenance(uuid), True)
        self.assertEqual(dag.get_node_data(uuid)._uuid, uuid)
        self.assertEqual(dag.get_node_data(uuid).type, 'Mapping')

    def test_artifact_with_collection_of_inputs(self):
        dag = self.das.merged_mappings.dag
        uuid = self.das.merged_mappings.uuid
        root_node = dag.get_node_data(uuid)
        self.assertEqual(root_node.type, 'Mapping')

        exp_parents = {self.das.mapping1.uuid, self.das.mapping2.uuid}
        self.assertEqual(dag.predecessors(uuid), exp_parents)

    def test_provdag_initialized_from_provdag(self):
        for dag in [self.das.single_int.dag, self.das.concated_ints.dag,
                    self.das.merged_mappings.dag]:
            copied = ProvDAG(dag)
            self.assertEqual(dag, copied)
            self.assertIsNot(dag, copied)

    def test_union_zero_or_one_dags(self):
        with self.assertRaisesRegex(ValueError, "pass.*two ProvDAGs"):
            ProvDAG.union([])

        with self.assertRaisesRegex(ValueError, "pass.*two ProvDAGs"):
            ProvDAG.union([self.das.single_int.dag])

    def test_union_identity(self):
        dag = self.das.single_int.dag
        uuid = self.das.single_int.uuid
        unioned_dag = ProvDAG.union([dag, dag])

        self.assertEqual(dag, unioned_dag)
        self.assertSetEqual({uuid}, unioned_dag._parsed_artifact_uuids)
        self.assertEqual(unioned_dag.provenance_is_valid, ValidationCode.VALID)
        self.assertRegex(
            repr(unioned_dag),
            f'ProvDAG representing the provenance.*Artifacts.*{uuid}'
        )

    def test_union_two(self):
        unioned_dag = ProvDAG.union(
            [self.das.single_int.dag, self.das.int_seq2.dag])

        self.assertEqual(
            {self.das.single_int.uuid, self.das.int_seq2.uuid},
            unioned_dag._parsed_artifact_uuids
        )
        self.assertEqual(unioned_dag.provenance_is_valid, ValidationCode.VALID)

        rep = repr(unioned_dag)
        self.assertRegex(
            rep, 'ProvDAG representing the provenance.*Artifacts.'
        )
        self.assertRegex(rep, f'{self.das.single_int.uuid}')
        self.assertRegex(rep, f'{self.das.int_seq2.uuid}')

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 2
        )

    def test_union_many(self):
        unioned_dag = ProvDAG.union([
            self.das.single_int.dag,
            self.das.int_seq1.dag,
            self.das.mapping1.dag
        ])
        self.assertEqual(
            {
                self.das.single_int.uuid,
                self.das.int_seq1.uuid,
                self.das.mapping1.uuid
            },
            unioned_dag._parsed_artifact_uuids
        )
        self.assertEqual(unioned_dag.provenance_is_valid, ValidationCode.VALID)

        rep = repr(unioned_dag)
        self.assertRegex(
            rep, 'ProvDAG representing the provenance.*Artifacts.*'
        )
        self.assertRegex(rep, f'{self.das.single_int.uuid}')
        self.assertRegex(rep, f'{self.das.int_seq1.uuid}')
        self.assertRegex(rep, f'{self.das.mapping1.uuid}')

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 3
        )

    def test_union_self_missing_checksums_md5(self):
        unioned_dag = ProvDAG.union(
            [self.das.dag_missing_md5, self.das.single_int.dag]
        )

        self.assertRegex(
            repr(unioned_dag),
            'ProvDAG representing the provenance.*Artifacts.*'
            f'{self.das.single_int.uuid}'
        )

        # The ChecksumDiff==None from the tinkered dag gets ignored...
        self.assertEqual(unioned_dag.checksum_diff, ChecksumDiff({}, {}, {}))

        # ...but this should make clear that the provenance is bad
        # (or that the user opted out of validation)
        self.assertEqual(
            unioned_dag.provenance_is_valid, ValidationCode.INVALID)

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

    def test_union_other_missing_checksums_md5(self):
        '''
        Tests unions of v5 dags where the other ProvDAG is missing its
        checksums.md5 but the calling ProvDAG is not
        '''
        unioned_dag = ProvDAG.union([self.das.single_int.dag,
                                     self.das.dag_missing_md5])

        self.assertRegex(repr(unioned_dag),
                         'ProvDAG representing the provenance.*Artifacts.*'
                         f'{self.das.single_int.uuid}')

        self.assertEqual(unioned_dag.checksum_diff, ChecksumDiff({}, {}, {}))

        self.assertEqual(
            unioned_dag.provenance_is_valid, ValidationCode.INVALID
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

    def test_union_both_missing_checksums_md5(self):
        '''
        Tests unions of v5 dags where both artifacts are missing their
        checksums.md5 files.
        '''
        unioned_dag = ProvDAG.union(
            [self.das.dag_missing_md5, self.das.dag_missing_md5])

        self.assertRegex(
            repr(unioned_dag),
            'ProvDAG representing the provenance.*Artifacts.*'
            f'{self.das.single_int.uuid}'
        )

        # Both DAGs have NoneType checksum_diffs, so the ChecksumDiff==None
        self.assertEqual(unioned_dag.checksum_diff, None)

        self.assertEqual(
            unioned_dag.provenance_is_valid, ValidationCode.INVALID
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

    def test_union_v0_v1_archives(self):
        unioned_dag = ProvDAG.union(
            [self.das.table_v0.dag, self.das.concated_ints_v2.dag]
        )

        self.assertIn(f'{self.das.table_v0.uuid}', repr(unioned_dag))
        self.assertIn(f'{self.das.concated_ints_v2.uuid}', repr(unioned_dag))

        self.assertEqual(
            unioned_dag.provenance_is_valid, ValidationCode.PREDATES_CHECKSUMS
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 2
        )

        self.assertFalse(
            unioned_dag.node_has_provenance(self.das.table_v0.uuid)
        )
        self.assertTrue(
            unioned_dag.node_has_provenance(self.das.concated_ints_v2.uuid)
        )

    def test_union_v3_v5_archives(self):
        unioned_dag = ProvDAG.union(
            [self.das.concated_ints_v3.dag, self.das.concated_ints_v5.dag]
        )

        self.assertIn(f'{self.das.concated_ints_v3.uuid}', repr(unioned_dag))
        self.assertIn(f'{self.das.concated_ints_v5.uuid}', repr(unioned_dag))

        self.assertEqual(
            unioned_dag.provenance_is_valid,
            ValidationCode.PREDATES_CHECKSUMS
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 2
        )

        self.assertTrue(
            unioned_dag.node_has_provenance(self.das.concated_ints_v3.uuid)
        )
        self.assertTrue(
            unioned_dag.node_has_provenance(self.das.concated_ints_v5.uuid)
        )

    def test_union_v5_v6_archives(self):
        unioned_dag = ProvDAG.union(
            [self.das.concated_ints_v5.dag, self.das.concated_ints_v6.dag]
        )

        self.assertIn(f'{self.das.concated_ints_v5.uuid}', repr(unioned_dag))
        self.assertIn(f'{self.das.concated_ints_v6.uuid}', repr(unioned_dag))

        self.assertEqual(
            unioned_dag.provenance_is_valid, ValidationCode.VALID
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 2
        )

        self.assertTrue(
            unioned_dag.node_has_provenance(self.das.concated_ints_v5.uuid)
        )
        self.assertTrue(
            unioned_dag.node_has_provenance(self.das.concated_ints_v6.uuid)
        )

    def test_dag_is_superset(self):
        '''
        Tests union of three dags, where one dag is a proper superset of the
        others. We expect three _parsed_artifact_uuids, one terminal uuid,
        and one weakly_connected_component.
        '''
        unioned_dag = ProvDAG.union([
            self.das.concated_ints.dag,
            self.das.int_seq1.dag,
            self.das.int_seq2.dag
        ])

        self.assertIn(
            self.das.int_seq1.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.int_seq2.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.concated_ints.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertEqual(len(unioned_dag._parsed_artifact_uuids), 3)

        self.assertEqual(len(unioned_dag.terminal_uuids), 1)
        self.assertEqual(
            unioned_dag.terminal_uuids, {self.das.concated_ints.uuid}
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

        # == tests identity of objects in memory, so we need is_isomorphic
        self.assertTrue(
            nx.is_isomorphic(self.das.concated_ints.dag.dag, unioned_dag.dag)
        )
        self.assertEqual(self.das.concated_ints.dag, unioned_dag)

    def test_three_artifacts_two_terminal_uuids(self):
        '''
        Tests union of four dags, where two artifacts have shared parents but
        no direct relationship with eachother. We expect four
        _parsed_artifact_uuids, two terminal uuids, and one
        weakly_connected_component.
        '''
        unioned_dag = ProvDAG.union([
            self.das.int_seq1.dag,
            self.das.int_seq2.dag,
            self.das.concated_ints.dag,
            self.das.other_concated_ints.dag
        ])

        self.assertIn(
            self.das.int_seq1.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.int_seq2.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.concated_ints.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.other_concated_ints.uuid,
            unioned_dag._parsed_artifact_uuids
        )
        self.assertEqual(len(unioned_dag._parsed_artifact_uuids), 4)

        self.assertEqual(len(unioned_dag.terminal_uuids), 2)
        self.assertEqual(
            unioned_dag.terminal_uuids,
            {self.das.concated_ints.uuid, self.das.other_concated_ints.uuid}
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

    def test_one_analysis_two_artifacts(self):
        '''
        In this set of test archives, both artifacts are derived from the same
        parents, so should produce one connected DAG even though we are
        missing the parent artifacts used to create them
        '''
        unioned_dag = ProvDAG.union([
            self.das.concated_ints.dag,
            self.das.other_concated_ints.dag
        ])

        self.assertIn(
            self.das.concated_ints.uuid, unioned_dag._parsed_artifact_uuids
        )
        self.assertIn(
            self.das.other_concated_ints.uuid,
            unioned_dag._parsed_artifact_uuids
        )
        self.assertEqual(len(unioned_dag._parsed_artifact_uuids), 2)

        self.assertEqual(len(unioned_dag.terminal_uuids), 2)
        self.assertEqual(
            unioned_dag.terminal_uuids,
            {self.das.concated_ints.uuid, self.das.other_concated_ints.uuid}
        )

        self.assertEqual(
            nx.number_weakly_connected_components(unioned_dag.dag), 1
        )

    def test_no_checksum_validation_on_intact_artifact(self):
        no_validation_dag = ProvDAG(
            self.das.concated_ints.filepath, validate_checksums=False
        )

        self.assertEqual(len(no_validation_dag.terminal_uuids), 1)
        terminal_uuid, *_ = no_validation_dag.terminal_uuids
        self.assertEqual(terminal_uuid, self.das.concated_ints.uuid)

        self.assertEqual(len(no_validation_dag), 3)
        self.assertEqual(
            no_validation_dag.provenance_is_valid,
            ValidationCode.VALIDATION_OPTOUT
        )
        self.assertEqual(no_validation_dag.checksum_diff, None)

    def test_no_checksum_validation_missing_checksums_md5(self):
        with generate_archive_with_file_removed(
            self.das.concated_ints.filepath,
            self.das.concated_ints.uuid,
            'checksums.md5'
        ) as altered_archive:
            dag = ProvDAG(altered_archive, validate_checksums=False)

            self.assertEqual(
                dag.provenance_is_valid, ValidationCode.VALIDATION_OPTOUT
            )

            self.assertEqual(dag.checksum_diff, None)

    def test_no_checksum_validation_missing_node_files(self):
        path_prefix = os.path.join('provenance', 'artifacts')
        root_uuid = self.das.concated_ints.uuid
        for removed_file in [
            'metadata.yaml',
            'citations.bib',
            'VERSION',
            'action/action.yaml'
        ]:
            for uuid in [self.das.int_seq1.uuid, self.das.int_seq2.uuid]:
                with generate_archive_with_file_removed(
                    self.das.concated_ints.filepath,
                    root_uuid,
                    os.path.join(path_prefix, uuid, removed_file)
                ) as altered_archive:
                    if removed_file == 'action/action.yaml':
                        file = 'action.yaml'
                    else:
                        file = removed_file

                    expected = (f'(?s)Malformed.*{file}.*{uuid}.*corrupt.*')
                    with self.assertRaisesRegex(ValueError, expected):
                        ProvDAG(altered_archive, validate_checksums=False)


class EmptyParserTests(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix='qiime2-test-parse-temp-')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_get_parser(self):
        parser = EmptyParser.get_parser(None)
        self.assertIsInstance(parser, EmptyParser)

    def test_get_parser_input_data_not_none(self):
        fn = 'not-a-zip.txt'
        fp = os.path.join(self.tempdir, fn)
        with open(fp, 'w') as fh:
            fh.write('some text\n')
        with self.assertRaisesRegex(
            TypeError, f"EmptyParser.*{fn} is not None"
        ):
            EmptyParser.get_parser(fp)

    def test_parse_a_nonetype(self):
        '''
        tests that we can actually create empty ProvDAGs
        '''
        parser = EmptyParser()
        parsed = parser.parse_prov(Config(), None)
        self.assertIsInstance(parsed, ParserResults)
        self.assertEqual(parsed.parsed_artifact_uuids, set())
        self.assertTrue(nx.is_isomorphic(parsed.prov_digraph, nx.DiGraph()))
        self.assertEqual(parsed.provenance_is_valid, ValidationCode.VALID)
        self.assertEqual(parsed.checksum_diff, None)


class ProvDAGParserTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_get_parser(self):
        for archive in self.das.all_artifact_versions:
            parser = ProvDAGParser.get_parser(archive.dag)
            self.assertIsInstance(parser, ProvDAGParser)

    def test_get_parser_input_data_not_a_provdag(self):
        fn = 'not_a_zip.txt'
        fp = os.path.join(self.das.tempdir, fn)
        with self.assertRaisesRegex(
                TypeError, f"ProvDAGParser.*{fn} is not a ProvDAG"):
            ProvDAGParser.get_parser(fp)

    def test_parse_a_provdag(self):
        parser = ProvDAGParser()
        for archive in self.das.all_artifact_versions:
            dag = archive.dag
            parsed = parser.parse_prov(Config(), dag)
            self.assertIsInstance(parsed, ParserResults)
            self.assertEqual(
                parsed.parsed_artifact_uuids, dag._parsed_artifact_uuids
            )
            self.assertTrue(nx.is_isomorphic(parsed.prov_digraph, dag.dag))
            self.assertEqual(
                parsed.provenance_is_valid, dag.provenance_is_valid
            )
            self.assertEqual(parsed.checksum_diff, dag.checksum_diff)


class SelectParserTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_correct_parser_type(self):
        empty = select_parser(None)
        self.assertIsInstance(empty, EmptyParser)

        archive = select_parser(self.das.concated_ints.filepath)
        self.assertIsInstance(archive, ArchiveParser)

        dag = ProvDAG()
        pdag = select_parser(dag)
        self.assertIsInstance(pdag, ProvDAGParser)

        # check dir_fp as fp
        test_dir = os.path.join(self.tempdir, 'parse_dir_test')
        os.mkdir(test_dir)
        dir_fp = pathlib.Path(test_dir)
        dir_p = select_parser(dir_fp)
        self.assertIsInstance(dir_p, DirectoryParser)

        # check dir_fp as str
        dir_fp_str = str(dir_fp)
        dir_p = select_parser(dir_fp_str)
        self.assertIsInstance(dir_p, DirectoryParser)

    def test_correct_archive_parser_version(self):
        parsers = [
            ParserV0, ParserV1, ParserV2, ParserV3, ParserV4, ParserV5,
            ParserV6
        ]
        for archive, parser in zip(self.das.all_artifact_versions, parsers):
            handler = select_parser(archive.filepath)
            self.assertEqual(type(handler), parser)


class ParseProvenanceTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.cfg = Config()

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_parse_with_artifact_parser(self):
        uuid = self.das.concated_ints.uuid
        fp = self.das.concated_ints.filepath
        parser_results = parse_provenance(self.cfg, fp)

        self.assertIsInstance(parser_results, ParserResults)
        p_a_uuids = parser_results.parsed_artifact_uuids
        self.assertIsInstance(p_a_uuids, set)
        self.assertIsInstance(next(iter(p_a_uuids)), str)
        self.assertEqual(len(parser_results.prov_digraph), 3)
        self.assertIn(uuid, parser_results.prov_digraph)
        self.assertIsInstance(
            parser_results.prov_digraph.nodes[uuid]['node_data'],
            ProvNode
        )
        self.assertEqual(
            parser_results.provenance_is_valid, ValidationCode.VALID
        )
        self.assertEqual(
            parser_results.checksum_diff, ChecksumDiff({}, {}, {})
        )

    def test_parse_with_provdag_parser(self):
        uuid = self.das.concated_ints.uuid
        dag = self.das.concated_ints.dag
        parser_results = parse_provenance(self.cfg, dag)

        self.assertIsInstance(parser_results, ParserResults)
        p_a_uuids = parser_results.parsed_artifact_uuids
        self.assertIsInstance(p_a_uuids, set)
        self.assertIsInstance(next(iter(p_a_uuids)), str)
        self.assertEqual(len(parser_results.prov_digraph), 3)
        self.assertIn(uuid, parser_results.prov_digraph)
        self.assertIsInstance(
            parser_results.prov_digraph.nodes[uuid]['node_data'],
            ProvNode
        )
        self.assertEqual(
            parser_results.provenance_is_valid, ValidationCode.VALID
        )
        self.assertEqual(
            parser_results.checksum_diff, ChecksumDiff({}, {}, {})
        )

    def test_parse_with_empty_parser(self):
        res = parse_provenance(self.cfg, None)
        self.assertIsInstance(res, ParserResults)
        self.assertEqual(res.parsed_artifact_uuids, set())
        self.assertTrue(nx.is_isomorphic(res.prov_digraph, nx.DiGraph()))
        self.assertEqual(res.provenance_is_valid, ValidationCode.VALID)
        self.assertEqual(res.checksum_diff, None)

    def test_parse_with_directory_parser(self):
        # Non-recursive
        parse_dir_fp = os.path.join(self.tempdir, 'parse_dir')
        os.mkdir(parse_dir_fp)
        concated_ints_path = os.path.join(parse_dir_fp, 'concated-ints.qza')
        shutil.copy(self.das.concated_ints.filepath, concated_ints_path)

        res = parse_provenance(self.cfg, parse_dir_fp)
        self.assertEqual(self.cfg.recurse, False)
        self.assertIsInstance(res, ParserResults)
        concated_ints_uuid = self.das.concated_ints.uuid
        self.assertEqual(res.parsed_artifact_uuids, {concated_ints_uuid})
        self.assertEqual(len(res.prov_digraph), 3)
        self.assertEqual(res.provenance_is_valid, ValidationCode.VALID)
        self.assertEqual(res.checksum_diff, ChecksumDiff({}, {}, {}))

        # Recursive
        inner_dir_path = os.path.join(parse_dir_fp, 'inner-dir')
        os.mkdir(inner_dir_path)
        mapping_path = os.path.join(inner_dir_path, 'mapping1.qza')
        shutil.copy(self.das.mapping1.filepath, mapping_path)

        self.cfg.recurse = True
        self.assertEqual(self.cfg.recurse, True)
        res = parse_provenance(self.cfg, parse_dir_fp)
        self.assertIsInstance(res, ParserResults)
        mapping_uuid = self.das.mapping1.uuid
        self.assertEqual(
            res.parsed_artifact_uuids, {concated_ints_uuid, mapping_uuid}
        )
        self.assertEqual(len(res.prov_digraph), 4)
        self.assertEqual(res.provenance_is_valid, ValidationCode.VALID)
        self.assertEqual(res.checksum_diff, ChecksumDiff({}, {}, {}))

    def test_parse_with_directory_parser_bad_dir_path(self):
        dir_fp = os.path.join(self.tempdir, 'fake_dir')
        with self.assertRaisesRegex(Exception, 'not a valid dir'):
            parse_provenance(self.cfg, dir_fp)

    def test_no_correct_parser_found_error(self):
        input_data = {'this': 'is not parseable'}
        with self.assertRaisesRegex(
            UnparseableDataError,
            f'(?s)Input data {input_data}.*not supported.*'
            'ArchiveParser expects a string or pathlib.PosixPath.*'
            'DirectoryParser.*expects a directory.*'
            'ProvDAGParser.*is not a ProvDAG.*'
            'EmptyParser.*is not None'
        ):
            select_parser(input_data)


class DirectoryParserTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir
        cls.cfg = Config()

        parse_dir_fp = os.path.join(cls.tempdir, 'parse-dir')
        os.mkdir(parse_dir_fp)
        concated_ints_path = os.path.join(parse_dir_fp, 'concated-ints.qza')
        shutil.copy(cls.das.concated_ints.filepath, concated_ints_path)
        int_seq_path = os.path.join(parse_dir_fp, 'int-seq1.qza')
        shutil.copy(cls.das.int_seq1.filepath, int_seq_path)

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_parse_empty_dir(self):
        empty_dir_path = os.path.join(self.tempdir, 'empty-dir')
        os.mkdir(empty_dir_path)

        with self.assertRaisesRegex(
            ValueError, f'No .qza or .qzv files.*{empty_dir_path}'
        ):
            ProvDAG(empty_dir_path)

    def test_directory_parser_works_regardless_trailing_slash(self):
        parse_dir = os.path.join(self.tempdir, 'parse-dir')
        dag = ProvDAG(parse_dir)
        dag2 = ProvDAG(parse_dir)
        self.assertEqual(dag, dag2)

    def test_directory_parser_captures_all_parsed_artifact_uuids(self):
        '''
        The test dir contains a concatenated ints artifact and an int sequence.
        Though the concatenated ints artifact contains the int sequence as one
        of its parents, both should be present in _parsed_artifact_uuids
        because both are considered terminal outputs of the analysis, because
        both are present in the parsed directory.
        '''
        parse_dir = os.path.join(self.tempdir, 'parse-dir')
        dag = ProvDAG(parse_dir, recurse=True)
        self.assertEqual(
            dag._parsed_artifact_uuids,
            {self.das.concated_ints.uuid, self.das.int_seq1.uuid}
        )

    def test_directory_parser_handles_duplicates(self):
        with tempfile.TemporaryDirectory() as tempdir:
            shutil.copy(self.das.concated_ints.filepath, tempdir)
            copy_path = os.path.join(tempdir, 'concated_ints_2.qza')
            shutil.copy(self.das.concated_ints.filepath, copy_path)

            both_dag = ProvDAG(tempdir)
            self.assertEqual(len(both_dag._parsed_artifact_uuids), 1)

            one_dag = ProvDAG(copy_path)
            self.assertEqual(one_dag, both_dag)

    def test_directory_parser_idempotent_with_parse_and_union(self):
        # Non-recursive
        concated_ints_path = os.path.join(
            self.tempdir, 'parse-dir', 'concated-ints.qza'
        )
        int_seq_path = os.path.join(self.tempdir, 'parse-dir', 'int-seq1.qza')
        concated_ints_dag = ProvDAG(concated_ints_path)
        int_seq_dag = ProvDAG(int_seq_path)
        union_dag = ProvDAG.union([concated_ints_dag, int_seq_dag])
        parse_dir_path = os.path.join(self.tempdir, 'parse-dir')
        dir_dag = ProvDAG(parse_dir_path)
        self.assertEqual(union_dag, dir_dag)

        # Recursive
        inner_dir = os.path.join(self.tempdir, 'parse-dir', 'inner-dir')
        os.mkdir(inner_dir)
        mapping_path = os.path.join(inner_dir, 'mapping1.qza')
        shutil.copy(self.das.mapping1.filepath, mapping_path)

        mapping_dag = ProvDAG(mapping_path)
        inner_dir_union_dag = ProvDAG.union([
            concated_ints_dag, int_seq_dag, mapping_dag
        ])
        recursive_dir_dag = ProvDAG(parse_dir_path, recurse=True)
        self.assertEqual(inner_dir_union_dag, recursive_dir_dag)

    def test_directory_parser_multiple_imports(self):
        outer_path = os.path.join(self.tempdir, 'mutliple-import-outer')
        os.mkdir(outer_path)
        inner_path = os.path.join(outer_path, 'inner')
        os.mkdir(inner_path)

        shutil.copy(
            self.das.mapping1.filepath,
            os.path.join(outer_path, 'mapping1.qza')
        )
        shutil.copy(
            self.das.mapping2.filepath,
            os.path.join(outer_path, 'mapping2.qza')
        )
        shutil.copy(
            self.das.mapping1.filepath,
            os.path.join(inner_path, 'duplicate-mapping1.qza')
        )
        shutil.copy(
            self.das.mapping2.filepath,
            os.path.join(inner_path, 'duplicate-mapping2.qza')
        )

        inner_dag = ProvDAG(inner_path)
        self.assertEqual(len(inner_dag), 2)
        self.assertIn(self.das.mapping1.uuid, inner_dag.dag)
        self.assertIn(self.das.mapping2.uuid, inner_dag.dag)

        outer_dag = ProvDAG(outer_path)
        self.assertEqual(len(inner_dag), 2)
        self.assertIn(self.das.mapping1.uuid, outer_dag.dag)
        self.assertIn(self.das.mapping2.uuid, outer_dag.dag)
        self.assertEqual(inner_dag, outer_dag)

        recursive_outer_dag = ProvDAG(outer_path, recurse=True)
        self.assertEqual(len(inner_dag), 2)
        self.assertIn(self.das.mapping1.uuid, recursive_outer_dag.dag)
        self.assertIn(self.das.mapping2.uuid, recursive_outer_dag.dag)
        self.assertEqual(inner_dag, outer_dag, recursive_outer_dag)

    def test_verbose(self):
        buffer = io.StringIO()
        concated_ints = 'parse-dir/concated-ints.qza'
        int_seq = 'parse-dir/int-seq1.qza'
        with redirect_stdout(buffer):
            dag = ProvDAG(
                os.path.join(self.tempdir, 'parse-dir'),
                verbose=True,
                recurse=True
            )

        self.assertEqual(dag.cfg.verbose, True)

        stdout_log = buffer.getvalue()
        self.assertRegex(stdout_log, f'parsing.*{concated_ints}')
        self.assertRegex(stdout_log, f'parsing.*{int_seq}')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
from datetime import datetime
from importlib.metadata import metadata
import pkg_resources
import textwrap
from typing import Any, Callable, List

from .parse import ProvDAG

from qiime2.sdk import Action
from qiime2.plugins import ArtifactAPIUsage
from qiime2.sdk.usage import (
    Usage, UsageVariable, UsageInputs, UsageOutputs
)


def build_header(
    shebang: str = '',
    boundary: str = '',
    copyright: str = '',
    extra_text: List[str] = []
) -> List[str]:
    '''
    Constructs the header contents for a replay script.

    Parameters
    ----------
    shebang : str
        The shebang line to add to the rendered script, if any.
    boundary : str
        The visual boundary to add to the rendred script, if any.
    copyright : str
        The copyright notice to add to the rendered script, if any.
    extra_text : list of str
        Extra lines of text to add to the header in the rendered script,
        if any.

    Returns
    -------
    list of str
        The constructed header lines.
    '''
    qiime2_md = metadata('qiime2')
    vzn = qiime2_md['Version']
    ts = datetime.now()
    header = []

    if shebang:
        header.append(shebang)

    if boundary:
        header.append(boundary)

    header.extend([
        f'# Auto-generated by qiime2 v.{vzn} at '
        f'{ts.strftime("%I:%M:%S %p")} on {ts.strftime("%d %b, %Y")}',
    ])

    if copyright:
        header.extend(copyright)

    header.append(
        '# For User Support, post to the QIIME2 Forum at '
        'https://forum.qiime2.org.'
    )

    if extra_text:
        header.extend(extra_text)
    if boundary:
        header.append(boundary)
    return header


def build_footer(dag: ProvDAG, boundary: str) -> List[str]:
    '''
    Constructs the footer contents for a replay script.

    Parameters
    ----------
    dag : ProvDAG
        The ProvDAG object representing the input artifact(s).
    boundary : str
        The visual boundary demarcating the footer.

    Returns
    -------
    list of str
        The constructed footer lines as a list of strings.
    '''
    footer = []
    pairs = []
    uuids = sorted(dag._parsed_artifact_uuids)
    # two UUIDs fit on a line
    for idx in range(0, u_len := len(uuids), 2):
        if idx == u_len - 1:
            pairs.append(f'# {uuids[idx]}')
        else:
            pairs.append(f'# {uuids[idx]} \t {uuids[idx + 1]}')

    footer.append(boundary)
    footer.append(
        '# The following QIIME 2 Results were parsed to produce this script:'
    )
    footer.extend(pairs)
    footer.append(boundary)
    footer.append('')
    return footer


class ReplayPythonUsage(ArtifactAPIUsage):
    shebang = '#!/usr/bin/env python'
    header_boundary = '# ' + ('-' * 77)
    copyright = pkg_resources.resource_string(
        'qiime2.core.archive.provenance_lib',
        'assets/copyright_note.txt'
    ).decode('utf-8').split('\n')
    how_to = pkg_resources.resource_string(
        'qiime2.core.archive.provenance_lib',
        'assets/python_howto.txt'
    ).decode('utf-8').split('\n')

    def __init__(
        self,
        enable_assertions: bool = False,
        action_collection_size: int = 2
    ):
        '''
        Identical to parent, but with smaller default action_collection_size.

        Parameters
        ----------
        enable_assertions : bool
            Whether to render has-line-matching and output type assertions.
        action_collection_size : int
            The maximum number of outputs returned by an action above which
            results are grouped into and destructured from a single variable.
        '''
        super().__init__()
        self.enable_assertions = enable_assertions
        self.action_collection_size = action_collection_size
        self._reset_state(reset_global_imports=True)

    def _reset_state(self, reset_global_imports=False):
        '''
        Clears all state associated with the usage driver, excepting global
        imports by default.

        Parameters
        ----------
        resest_global_imports : bool
            Whether to reset self.global_imports to an empty set.
        '''
        self.local_imports = set()
        self.header = []
        self.recorder = []
        self.footer = []
        self.init_data_refs = dict()
        if reset_global_imports:
            self.global_imports = set()

    def _template_action(
        self, action: Action, input_opts: UsageInputs, variables: UsageOutputs
    ):
        '''
        Templates the artifact api python code for the action `action`.

        Extends the parent method to:
        - accommodate action signatures that may differ between those found in
          provenance and those accessible in the currently executing
          environment
        - render artifact api code that saves the results to disk.

        Parameters
        ----------
        action : Action
            The qiime2 Action object.
        input_opts : UsageInputs
            The UsageInputs mapping for the action.
        variables : UsageOutputs
            The UsageOutputs object for the action.
        '''
        action_f = action.get_action()
        if (
            len(variables) > self.action_collection_size
            or len(action_f.signature.outputs) > 5
        ):
            output_vars = 'action_results'
        else:
            output_vars = self._template_outputs(action, variables)

        plugin_id = action.plugin_id
        action_id = action.action_id
        lines = [
            f'{output_vars} = {plugin_id}_actions.{action_id}('
        ]

        all_inputs = (list(action_f.signature.inputs.keys()) +
                      list(action_f.signature.parameters.keys()))
        for k, v in input_opts.items():
            line = ''
            if k not in all_inputs:
                line = self.INDENT + (
                    '# FIXME: The following parameter name was not found in '
                    'your current\n    # QIIME 2 environment. This may occur '
                    'when the plugin version you have\n    # installed does '
                    'not match the version used in the original analysis.\n '
                    ' # Please see the docs and correct the parameter name '
                    'before running.\n'
                )
            line += self._template_input(k, v)
            lines.append(line)

        lines.append(')')

        if (
            len(variables) > self.action_collection_size
            or len(action.get_action().signature.outputs) > 5
        ):
            for k, v in variables._asdict().items():
                interface_name = v.to_interface_name()
                lines.append('%s = action_results.%s' % (interface_name, k))

        lines.append(
            '# SAVE: comment out the following with \'# \' to skip saving '
            'Results to disk'
        )
        for k, v in variables._asdict().items():
            interface_name = v.to_interface_name()
            lines.append(
                '%s.save(\'%s\')' % (interface_name, interface_name,))

        lines.append('')
        self._add(lines)

    def _template_outputs(
        self, action: Action, variables: UsageOutputs
    ) -> str:
        '''
        Extends the parent method to allow the replay an action when a
        ProvDAG doesn't have a record of all outputs from an action. These
        unknown outputs are given the conventional '_' variable name.

        Parameters
        ----------
        action : Action
            The Action object associated with the output variables.
        variables : UsageOutputs
            The UsageOutputs object associated with the action.

        Returns
        -------
        str
            The templated output variables names as a comma-separated string.
        '''
        output_vars = []
        action_f = action.get_action()

        # need to coax the outputs into the correct order for unpacking
        for output in action_f.signature.outputs:
            try:
                variable = getattr(variables, output)
                output_vars.append(str(variable.to_interface_name()))
            except AttributeError:
                output_vars.append('_')

        if len(output_vars) == 1:
            output_vars.append('')

        return ', '.join(output_vars).strip()

    def init_metadata(
        self, name: str, factory: Callable, dumped_md_fn: str = ''
    ) -> UsageVariable:
        '''
        Renders the loading of Metadata from disk.

        Parameters
        ----------
        name : str
            The name of the created and returned UsageVariable.
        factory : Callable
            The factory responsible for constructing the metadata
            UsageVariable.
        dumped_md_fn : str
            Optional. The filename of the dumped metadata.

        Returns
        -------
        UsageVariable
            The UsageVariable of var_type metadata corresponding to the
            loaded metadata.
        '''
        var = super().init_metadata(name, factory)
        self._update_imports(from_='qiime2', import_='Metadata')
        input_fp = var.to_interface_name()
        if dumped_md_fn:
            lines = [f'{input_fp} = Metadata.load("{dumped_md_fn}.tsv")']
        else:
            self.comment(
                'NOTE: You may substitute already-loaded Metadata for the '
                'following, or cast a pandas.DataFrame to Metadata as needed.'
            )
            lines = [f'{input_fp} = Metadata.load(<your metadata filepath>)']

        self._add(lines)
        return var

    def import_from_format(
        self,
        name: str,
        semantic_type: str,
        variable: UsageVariable,
        view_type: Any = None
    ):
        '''
        Extends the parent method to:
        - use '<your data here>' instead of import_fp for the import filepath
        - render artifact api code that saves the result to disk.

        Parameters
        ----------
        name : str
            The name of the UsageVariable `variable`.
        semantic_type : str
            The semantic type of the UsageVariable `variable`.
        variable : UsageVariable
            The usage variable object.
        view_type : str or some format
            The view type to use for importing.

        Returns
        -------
        UsageVariable
            The imported artifact UsageVariable.
        '''
        imported_var = Usage.import_from_format(
            self, name, semantic_type, variable, view_type=view_type
        )

        interface_name = imported_var.to_interface_name()
        import_fp = self.repr_raw_variable_name('<your data here>')

        lines = [
            '%s = Artifact.import_data(' % (interface_name,),
            self.INDENT + '%r,' % (semantic_type,),
            self.INDENT + '%r,' % (import_fp,),
        ]

        if view_type is not None:
            if type(view_type) is not str:
                # Show users where these formats come from when used in the
                # Python API to make things less 'magical'.
                import_path = super()._canonical_module(view_type)
                view_type = view_type.__name__
                if import_path is not None:
                    self._update_imports(from_=import_path,
                                         import_=view_type)
                else:
                    # May be in scope already, but something is quite wrong at
                    # this point, so assume the plugin_manager is sufficiently
                    # informed.
                    view_type = repr(view_type)
            else:
                view_type = repr(view_type)

            lines.append(self.INDENT + '%s,' % (view_type,))

        lines.extend([
            ')',
            '# SAVE: comment out the following with \'# \' to skip saving this'
            ' Result to disk',
            '%s.save(\'%s\')' % (interface_name, interface_name,),
            ''
        ])

        self._update_imports(from_='qiime2', import_='Artifact')
        self._add(lines)

        return imported_var

    class repr_raw_variable_name:
        # allows us to repr col name without enclosing quotes
        # (as in qiime2.qiime2.plugins.ArtifactAPIUsageVariable)
        def __init__(self, value):
            self.value = value

        def __repr__(self):
            return self.value

    def comment(self, line: str):
        '''
        Communicate that a comment should be rendered.

        Parameters
        ----------
        line : str
            The comment to be rendered.
        '''
        LINE_LEN = 79
        lines = textwrap.wrap(
            line,
            LINE_LEN,
            break_long_words=False,
            initial_indent='# ',
            subsequent_indent='# '
        )
        lines.append('')
        self._add(lines)

    def render(self, flush: bool = False) -> str:
        '''
        Return a newline-seperated string of Artifact API python code.

        Parameters
        ----------
        flush : bool
            Whether to 'flush' the current code. Importantly, this will clear
            the top-line imports for future invocations.

        Returns
        -------
        str
            The rendered string of python code.
        '''
        sorted_imps = sorted(self.local_imports)
        if self.header:
            self.header = self.header + ['']
        if self.footer:
            self.footer = [''] + self.footer
        if sorted_imps:
            sorted_imps = sorted_imps + ['']
        rendered = '\n'.join(
            self.header + sorted_imps + self.recorder + self.footer
        )
        if flush:
            self._reset_state()
        return rendered

    def build_header(self):
        '''Constructs a renderable header from its components.'''
        self.header.extend(build_header(
            self.shebang, self.header_boundary, self.copyright, self.how_to
        ))

    def build_footer(self, dag: ProvDAG):
        '''
        Constructs a renderable footer using the terminal uuids of a ProvDAG.
        '''
        self.footer.extend(build_footer(dag, self.header_boundary))



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
from datetime import timedelta
import os
import networkx as nx
import pathlib
from unittest.mock import MagicMock
import pandas as pd
import tempfile
import unittest
import zipfile

import pytest

from .._checksum_validator import ChecksumDiff, ValidationCode
from .testing_utilities import (
    DummyArtifacts, is_root_provnode_data, write_zip_archive
)
from ..archive_parser import (
    ProvNode, Config, _Action, _Citations, _ResultMetadata, ParserResults,
    ArchiveParser, ParserV0, ParserV1, ParserV2, ParserV3, ParserV4, ParserV5,
    ParserV6,
)
from ...provenance import MetadataInfo

from qiime2.core.testing.util import ReallyEqualMixin


class ParserVxTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_parse_root_md(self):
        for artifact in self.das.all_artifact_versions:
            fp = artifact.filepath
            uuid = artifact.uuid
            parser = ArchiveParser.get_parser(fp)
            with zipfile.ZipFile(fp) as zf:
                root_md = parser._parse_root_md(zf, uuid)
                self.assertEqual(root_md.uuid, uuid)
                if artifact == self.das.table_v0:
                    self.assertEqual(root_md.format, 'BIOMV210DirFmt')
                    self.assertEqual(root_md.type, 'FeatureTable[Frequency]')
                else:
                    self.assertEqual(root_md.format,
                                     'IntSequenceDirectoryFormat')
                    self.assertEqual(root_md.type, 'IntSequence1')

    def test_parse_root_md_no_md_yaml(self):
        for artifact in self.das.all_artifact_versions:
            parser = ArchiveParser.get_parser(artifact.filepath)

            with tempfile.TemporaryDirectory() as tempdir:
                with zipfile.ZipFile(artifact.filepath) as zf:
                    zf.extractall(tempdir)

                metadata_path = os.path.join(tempdir, artifact.uuid,
                                             'metadata.yaml')
                os.remove(metadata_path)
                fn = os.path.basename(artifact.filepath)
                fp = os.path.join(tempdir, fn)
                write_zip_archive(fp, tempdir)

                with zipfile.ZipFile(fp) as zf:
                    with self.assertRaisesRegex(
                        ValueError,
                        'Malformed.*metadata'
                    ):
                        parser._parse_root_md(zf, artifact.uuid)

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_populate_archive(self):
        for artifact in self.das.all_artifact_versions:
            parser = ArchiveParser.get_parser(artifact.filepath)
            fp = artifact.filepath
            uuid = artifact.uuid
            version = artifact.archive_version

            if version == 0:
                with self.assertWarnsRegex(
                    UserWarning,
                    'Artifact .*prior to provenance'
                ):
                    res = parser.parse_prov(Config(), fp)

            else:
                res = parser.parse_prov(Config(), fp)
                self.assertIsInstance(res, ParserResults)
                pa_uuids = res.parsed_artifact_uuids
                self.assertIsInstance(pa_uuids, set)
                self.assertIsInstance(next(iter(pa_uuids)), str)
                self.assertIsInstance(res.prov_digraph,
                                      (type(None), nx.DiGraph))
                self.assertIsInstance(res.provenance_is_valid, ValidationCode)

                if version < 5:
                    self.assertIsInstance(res.checksum_diff, type(None))
                else:
                    self.assertIsInstance(res.checksum_diff, ChecksumDiff)

                self.assertIn(uuid, res.prov_digraph)
                self.assertIsInstance(
                    res.prov_digraph.nodes[uuid]['node_data'], ProvNode)

    def test_validate_checksums(self):
        for artifact in self.das.all_artifact_versions:
            parser = ArchiveParser.get_parser(artifact.filepath)
            with zipfile.ZipFile(artifact.filepath) as zf:
                is_valid, diff = parser._validate_checksums(zf)
                if artifact.archive_version < 5:
                    self.assertEqual(is_valid,
                                     ValidationCode.PREDATES_CHECKSUMS)
                    self.assertEqual(diff, None)
                else:
                    self.assertEqual(is_valid, ValidationCode.VALID)
                    self.assertEqual(diff, ChecksumDiff({}, {}, {}))

    @pytest.mark.filterwarnings('ignore::UserWarning')
    def test_correct_validate_checksums_method_called(self):
        '''
        We want to confirm that parse_prov uses the local _validate_checksums
        even when it calls super().parse_prov() internally
        '''
        for artifact in self.das.all_artifact_versions:
            parser = ArchiveParser.get_parser(artifact.filepath)
            if artifact.archive_version < 5:
                parser._validate_checksums = MagicMock(
                    # return values only here to facilitate normal execution
                    return_value=(ValidationCode.PREDATES_CHECKSUMS, None)
                )
                parser.parse_prov(Config(), artifact.filepath)
                parser._validate_checksums.assert_called_once()
            else:
                parser._validate_checksums = MagicMock(
                    return_value=(
                        ValidationCode.VALID,
                        ChecksumDiff({}, {}, {})
                    )
                )
                parser.parse_prov(Config(), artifact.filepath)
                parser._validate_checksums.assert_called_once()


class ArchiveParserTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_get_parser(self):
        parsers = [
            ParserV0, ParserV1, ParserV2, ParserV3, ParserV4, ParserV5,
            ParserV6
        ]
        for artifact, parser_version in zip(
            self.das.all_artifact_versions, parsers
        ):
            parser = ArchiveParser.get_parser(artifact.filepath)
            self.assertEqual(type(parser), parser_version)

    def test_get_parser_nonexistent_fp(self):
        fn = 'not_a_filepath.qza'
        fp = os.path.join(self.tempdir, fn)
        with self.assertRaises(FileNotFoundError):
            ArchiveParser.get_parser(fp)

    def test_artifact_parser_parse_prov(self):
        with self.assertRaisesRegex(NotImplementedError, "Use a subclass"):
            ArchiveParser().parse_prov(Config(), 'doesnotmatter.txt')


class ResultMetadataTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

        cls.uuid = cls.das.concated_ints.uuid
        md_fp = f'{cls.uuid}/provenance/metadata.yaml'
        with zipfile.ZipFile(cls.das.concated_ints.filepath) as zf:
            cls.root_md = _ResultMetadata(zf, md_fp)

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_smoke(self):
        self.assertEqual(self.root_md.uuid, self.uuid)
        self.assertEqual(self.root_md.type, 'IntSequence1')
        self.assertEqual(self.root_md.format, 'IntSequenceDirectoryFormat')

    def test_repr(self):
        exp = (f'UUID:\t\t{self.uuid}\n'
               'Type:\t\tIntSequence1\n'
               'Data Format:\tIntSequenceDirectoryFormat')
        self.assertEqual(repr(self.root_md), exp)


class ActionTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

        action_path = os.path.join(cls.das.concated_ints_v6.uuid, 'provenance',
                                   'action', 'action.yaml')
        with zipfile.ZipFile(cls.das.concated_ints_v6.filepath) as zf:
            cls.concat_action = _Action(zf, action_path)

        action_path = os.path.join(cls.das.single_int.uuid, 'provenance',
                                   'action', 'action.yaml')
        with zipfile.ZipFile(cls.das.single_int.filepath) as zf:
            cls.import_action = _Action(zf, action_path)

        action_path = os.path.join(cls.das.pipeline_viz.uuid, 'provenance',
                                   'action', 'action.yaml')
        with zipfile.ZipFile(cls.das.pipeline_viz.filepath) as zf:
            cls.pipeline_action = _Action(zf, action_path)

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_action_id(self):
        exp = '5035a60e-6f9a-40d4-b412-48ae52255bb5'
        self.assertEqual(self.concat_action.action_id, exp)

    def test_action_type(self):
        self.assertEqual(self.concat_action.action_type, 'method')
        self.assertEqual(self.import_action.action_type, 'import')
        self.assertEqual(self.pipeline_action.action_type, 'pipeline')

    def test_runtime(self):
        exp_t = timedelta
        exp = timedelta(microseconds=6840)
        self.assertIsInstance(self.concat_action.runtime, exp_t)
        self.assertEqual(self.concat_action.runtime, exp)

    def test_runtime_str(self):
        exp = '6840 microseconds'
        self.assertEqual(self.concat_action.runtime_str, exp)

    def test_action(self):
        exp = 'concatenate_ints'
        self.assertEqual(self.concat_action.action_name, exp)

    def test_plugin(self):
        exp = 'dummy_plugin'
        self.assertEqual(self.concat_action.plugin, exp)

    '''
    Import is not handled by a plugin, so the parser provides values
    for the action_name and plugin properties not present in action.yaml
    '''
    def test_action_for_import_node(self):
        exp = 'import'
        self.assertEqual(self.import_action.action_name, exp)

    def test_plugin_for_import_node(self):
        exp = 'framework'
        self.assertEqual(self.import_action.plugin, exp)

    def test_inputs(self):
        exp = {
            'ints1': '8dea2f1a-2164-4a85-9f7d-e0641b1db22b',
            'ints2': '8dea2f1a-2164-4a85-9f7d-e0641b1db22b',
            'ints3': '7727c060-5384-445d-b007-b64b41a090ee'
        }
        self.assertEqual(self.concat_action.inputs, exp)

        exp = {}
        self.assertEqual(self.import_action.inputs, exp)

    def test_parameters(self):
        exp = {
            'int1': 7,
            'int2': 100,
        }
        self.assertEqual(self.concat_action.parameters, exp)

        exp = {}
        self.assertEqual(self.import_action.parameters, exp)

    def test_output_name(self):
        exp = 'concatenated_ints'
        self.assertEqual(self.concat_action.output_name, exp)

        exp = None
        self.assertEqual(self.import_action.output_name, exp)

    def test_format(self):
        exp = None
        self.assertEqual(self.concat_action.format, exp)
        self.assertEqual(self.import_action.format, exp)

    def test_transformers(self):
        int_seq_dir_citation = (
            'view|dummy-plugin:0.0.0-dev|IntSequenceDirectoryFormat|0'
        )
        transformer_citation = (
            'transformer|dummy-plugin:0.0.0-dev|builtins:list'
            '->IntSequenceDirectoryFormat|0'
        )
        output_citations = [transformer_citation, int_seq_dir_citation]

        exp = {
            'inputs': {
                'ints1':
                    [{
                        'from': 'IntSequenceDirectoryFormat',
                        'to': 'builtins:list',
                        'plugin': 'dummy-plugin',
                        'citations': [int_seq_dir_citation]
                    }],
                'ints2':
                    [{
                        'from': 'IntSequenceDirectoryFormat',
                        'to': 'builtins:list',
                        'plugin': 'dummy-plugin',
                        'citations': [int_seq_dir_citation]
                    }],
                'ints3':
                    [{
                        'from': 'IntSequenceV2DirectoryFormat',
                        'to': 'builtins:list',
                        'plugin': 'dummy-plugin',
                    }],
            },
            'output': [{
                'from': 'builtins:list',
                'to': 'IntSequenceDirectoryFormat',
                'plugin': 'dummy-plugin',
                'citations': output_citations
            }]
        }
        self.assertEqual(self.concat_action.transformers, exp)

    def test_repr(self):
        exp = (
            '_Action(action_id=5035a60e-6f9a-40d4-b412-48ae52255bb5, '
            'type=method, plugin=dummy_plugin, '
            'action=concatenate_ints)'
        )
        self.assertEqual(repr(self.concat_action), exp)


class CitationsTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

        cite_strs = ['cite_none', 'cite_one', 'cite_many']
        cls.bibs = [bib+'.bib' for bib in cite_strs]
        cls.zips = [
            os.path.join(cls.das.datadir, bib+'.zip') for bib in cite_strs
        ]

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_empty_bib(self):
        with zipfile.ZipFile(self.zips[0]) as zf:
            citations = _Citations(zf, self.bibs[0])
            self.assertEqual(len(citations.citations), 0)

    def test_citation(self):
        with zipfile.ZipFile(self.zips[1]) as zf:
            exp = 'framework'
            citations = _Citations(zf, self.bibs[1])
            for key in citations.citations:
                self.assertRegex(key, exp)

    def test_many_citations(self):
        exp = ['2020.6.0.dev0', 'unweighted_unifrac.+0',
               'unweighted_unifrac.+1', 'unweighted_unifrac.+2',
               'unweighted_unifrac.+3', 'unweighted_unifrac.+4',
               'BIOMV210DirFmt', 'BIOMV210Format']
        with zipfile.ZipFile(self.zips[2]) as zf:
            citations = _Citations(zf, self.bibs[2])
            for i, key in enumerate(citations.citations):
                self.assertRegex(key, exp[i])

    def test_repr(self):
        exp = ("Citations(['framework|qiime2:2020.6.0.dev0|0'])")
        with zipfile.ZipFile(self.zips[1]) as zf:
            citations = _Citations(zf, self.bibs[1])
            self.assertEqual(repr(citations), exp)


class ProvNodeTests(unittest.TestCase, ReallyEqualMixin):
    @classmethod
    def setUpClass(cls):
        cls.das = DummyArtifacts()
        cls.tempdir = cls.das.tempdir

        # build root nodes for all archive format versions
        cfg = Config(parse_study_metadata=True)
        cls.nodes = {}
        for artifact in cls.das.all_artifact_versions:
            with zipfile.ZipFile(artifact.filepath) as zf:
                all_filenames = zf.namelist()
                root_md_fnames = filter(is_root_provnode_data, all_filenames)
                root_md_fps = [pathlib.Path(fp) for fp in root_md_fnames]
                cls.nodes[str(artifact.archive_version)] = \
                    ProvNode(cfg, zf, root_md_fps)

        with zipfile.ZipFile(cls.das.concated_ints_with_md.filepath) as zf:
            root_node_id = cls.das.concated_ints_with_md.uuid
            all_filenames = zf.namelist()
            dag = cls.das.concated_ints_with_md.dag
            for node in dag.nodes:
                md_path = os.path.join(
                    root_node_id, 'provenance', 'artifacts', node, 'action',
                    'metadata.tsv'
                )
                if md_path in all_filenames:
                    md_node_id = node
                else:
                    non_md_node_id = node

            # build a nonroot node without study metadata
            node_fps = [
                pathlib.Path(fp) for fp in all_filenames if
                non_md_node_id in fp and
                ('metadata.yaml' in fp or 'action.yaml' in fp
                 or 'VERSION' in fp)
            ]
            cls.nonroot_non_md_node = ProvNode(cfg, zf, node_fps)

            # build a nonroot node with study metadata
            all_filenames = zf.namelist()
            node_fps = [
                pathlib.Path(fp) for fp in all_filenames if
                md_node_id in fp and
                ('metadata.yaml' in fp or 'action.yaml' in fp
                 or 'VERSION' in fp)
            ]
            cls.nonroot_md_node = ProvNode(cfg, zf, node_fps)

            # build a root node and parse study metadata files
            root_md_fnames = filter(is_root_provnode_data, all_filenames)
            root_md_fps = [pathlib.Path(fp) for fp in root_md_fnames]
            cfg = Config(parse_study_metadata=True)
            cls.root_node_parse_md = ProvNode(cfg, zf, root_md_fps)

            # build a root node and don't parse study metadata files
            cfg = Config(parse_study_metadata=False)
            cls.root_node_dont_parse_md = ProvNode(cfg, zf, root_md_fps)

        # build a node with a collection as input
        with zipfile.ZipFile(cls.das.int_from_collection.filepath) as zf:
            all_filenames = zf.namelist()
            root_md_fnames = filter(is_root_provnode_data, all_filenames)
            root_md_fps = [pathlib.Path(fp) for fp in root_md_fnames]
            cls.input_collection_node = ProvNode(cfg, zf, root_md_fps)

        # build a node with an optional input that defaults to None
        with zipfile.ZipFile(cls.das.int_seq_optional_input.filepath) as zf:
            all_filenames = zf.namelist()
            root_md_fnames = filter(is_root_provnode_data, all_filenames)
            root_md_fps = [pathlib.Path(fp) for fp in root_md_fnames]
            cls.optional_input_node = ProvNode(cfg, zf, root_md_fps)

    @classmethod
    def tearDownClass(cls):
        cls.das.free()

    def test_smoke(self):
        self.assertTrue(True)
        for node_vzn in self.nodes:
            self.assertIsInstance(self.nodes[node_vzn], ProvNode)

    def test_node_properties(self):
        # hardcoded from test data
        framework_versions = {
            '0': '2.0.5',
            '1': '2017.2.0',
            '2': '2017.9.0',
            '3': '2018.2.0',
            '4': '2018.6.0',
            '5': '2018.11.0',
            '6': '2023.5.1',
        }

        for node, archive_version in zip(
            self.nodes, [str(i) for i in range(7)]
        ):
            if archive_version == '0':
                self.assertEqual(self.nodes[node].format, 'BIOMV210DirFmt')
                self.assertEqual(self.nodes[node].type,
                                 'FeatureTable[Frequency]')
                self.assertEqual(self.nodes[node].has_provenance, False)
            else:
                self.assertEqual(self.nodes[node].format,
                                 'IntSequenceDirectoryFormat')
                self.assertEqual(self.nodes[node].type, 'IntSequence1')
                if archive_version == '1':
                    self.assertEqual(self.nodes[node].has_provenance, False)
                else:
                    self.assertEqual(self.nodes[node].has_provenance, True)

            self.assertEqual(self.nodes[node].archive_version, archive_version)
            self.assertEqual(
                self.nodes[node].framework_version,
                framework_versions[archive_version]
            )

    def test_self_eq(self):
        self.assertReallyEqual(self.nodes['5'], self.nodes['5'])

    def test_eq(self):
        # Mock has no matching UUID
        mock_node = MagicMock()
        self.assertNotEqual(self.nodes['5'], mock_node)

        # Mock has bad UUID
        mock_node._uuid = 'gerbil'
        self.assertReallyNotEqual(self.nodes['5'], mock_node)

        # Matching UUIDs insufficient if classes differ
        mock_node._uuid = self.das.concated_ints_v5.uuid
        self.assertReallyNotEqual(self.nodes['5'], mock_node)
        mock_node.__class__ = ProvNode
        self.assertReallyEqual(self.nodes['5'], mock_node)

    def test_is_hashable(self):
        exp_hash = hash(self.das.concated_ints_v5.uuid)
        self.assertReallyEqual(hash(self.nodes['5']), exp_hash)

    def test_str(self):
        for node_vzn, artifact in zip(
            self.nodes, self.das.all_artifact_versions
        ):
            uuid = artifact.uuid
            self.assertRegex(str(self.nodes[node_vzn]),
                             f'(?s)UUID:\t\t{uuid}.*Type.*Data Format')

    def test_repr(self):
        for node_vzn, artifact in zip(
            self.nodes, self.das.all_artifact_versions
        ):
            uuid = artifact.uuid
            self.assertRegex(repr(self.nodes[node_vzn]),
                             f'(?s)UUID:\t\t{uuid}.*Type.*Data Format')

    def test_get_metadata_from_action(self):
        find_md = self.root_node_parse_md._get_metadata_from_Action

        # create dummy hash '0', not relevant here
        md = MetadataInfo(
            ['d5b4cf78-f5e2-44e0-aa24-66b02564e9f1'], 'metadata.tsv', '0'
        )
        action_details = {
            'parameters': [{'metadata': md}]
        }
        all_md, artifacts_as_md = find_md(action_details)
        all_exp = {'metadata': 'metadata.tsv'}
        a_as_md_exp = [{
            'artifact_passed_as_metadata':
            'd5b4cf78-f5e2-44e0-aa24-66b02564e9f1'
        }]
        self.assertEqual(all_md, all_exp)
        self.assertEqual(artifacts_as_md, a_as_md_exp)

    def test_get_metadata_from_action_with_no_params(self):
        find_md = self.nodes['5']._get_metadata_from_Action
        action_details = \
            {'parameters': []}
        all_md, artifacts_as_md = find_md(action_details)
        self.assertEqual(all_md, {})
        self.assertEqual(artifacts_as_md, [])

        action_details = {'non-parameters-key': 'here is a thing'}
        all_md, artifacts_as_md = find_md(action_details)
        self.assertEqual(all_md, {})
        self.assertEqual(artifacts_as_md, [])

    def test_metadata_available_in_property(self):
        self.assertEqual(type(self.nonroot_md_node.metadata), dict)
        self.assertIn('metadata', self.nonroot_md_node.metadata)
        self.assertEqual(type(self.nonroot_md_node.metadata['metadata']),
                         pd.DataFrame)

    def test_metadata_not_available_in_property_w_opt_out(self):
        self.assertEqual(self.root_node_dont_parse_md.metadata, None)

    def test_metadata_is_correct(self):
        self.assertIn('metadata', self.nonroot_md_node.metadata)

        md_data = {
            'id': ['#q2:types', '0'],
            'a': ['categorical', '42'],
        }
        md_exp = pd.DataFrame(md_data, columns=md_data.keys())
        pd.testing.assert_frame_equal(
            md_exp, self.nonroot_md_node.metadata['metadata']
        )

    def test_has_no_provenance_so_no_metadata(self):
        self.assertEqual(self.nodes['0'].has_provenance, False)
        self.assertEqual(self.nodes['0'].metadata, None)

    def test_node_has_provenance_but_no_metadata(self):
        self.assertEqual(self.nonroot_non_md_node.has_provenance, True)
        self.assertEqual(self.nonroot_non_md_node.metadata, {})

    def test_parse_metadata_for_nonroot_node(self):
        self.assertEqual(self.nonroot_md_node.has_provenance, True)
        self.assertIn('metadata', self.nonroot_md_node.metadata)

    def test_parents(self):
        actual_parent_names = []
        for parent in self.nodes['5']._parents:
            actual_parent_names += parent.keys()

        self.assertIn('ints1', actual_parent_names)
        self.assertIn('ints2', actual_parent_names)
        self.assertIn('ints3', actual_parent_names)

        self.assertEqual(len(self.nodes['5']._parents), 3)
        self.assertEqual(len(actual_parent_names), 3)

    def test_parents_no_prov(self):
        no_prov_node = self.nodes['0']
        self.assertFalse(no_prov_node.has_provenance)
        self.assertEqual(no_prov_node._parents, None)

    def test_parents_with_artifact_passed_as_md(self):
        actual_parent_names = []
        for parent in self.nonroot_md_node._parents:
            actual_parent_names += parent.keys()

        self.assertIn('ints', actual_parent_names)
        self.assertIn('artifact_passed_as_metadata', actual_parent_names)

        self.assertEqual(len(self.nonroot_md_node._parents), 2)
        self.assertEqual(len(actual_parent_names), 2)

    def test_parents_for_import_node(self):
        uuid = self.das.single_int.uuid
        with zipfile.ZipFile(self.das.single_int.filepath) as zf:
            required_fps = ('VERSION', 'metadata.yaml', 'action.yaml')
            import_node_fps = [
                pathlib.Path(fp) for fp in zf.namelist()
                if uuid in fp
                and any(map(lambda x: x in fp, required_fps))
            ]
            import_node = ProvNode(Config(), zf, import_node_fps)

        self.assertEqual(import_node._parents, [])

    def test_parents_collection_of_inputs(self):
        parents = self.input_collection_node._parents
        self.assertIn('int1', parents[0].keys())
        self.assertIn('int2', parents[1].keys())
        self.assertEqual(len(parents), 2)

    def test_parents_optional_input(self):
        # NOTE: The None-type input is not captured
        parents = self.optional_input_node._parents
        self.assertIn('ints', parents[0].keys())
        self.assertEqual(len(parents), 1)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from .type import SingleInt, Mapping
from qiime2.sdk.result import ResultCollection
from qiime2.core.testing.util import PipelineError


def parameter_only_pipeline(ctx, int1, int2=2, metadata=None):
    identity_with_optional_metadata = ctx.get_action(
        'dummy_plugin', 'identity_with_optional_metadata')
    concatenate_ints = ctx.get_action('dummy_plugin', 'concatenate_ints')

    ints1 = ctx.make_artifact('IntSequence2', [int1, int2, 3])
    ints2, = identity_with_optional_metadata(ints1, metadata)
    ints3, = identity_with_optional_metadata(ints1, metadata)
    more_ints, = concatenate_ints(ints3, ints2, ints1, int1=int1, int2=int2)

    return ints1, more_ints


def typical_pipeline(ctx, int_sequence, mapping, do_extra_thing, add=1):
    split_ints = ctx.get_action('dummy_plugin', 'split_ints')
    most_common_viz = ctx.get_action('dummy_plugin', 'most_common_viz')

    left, right = split_ints(int_sequence)
    if do_extra_thing:
        left = ctx.make_artifact(
            'IntSequence1', [i + add for i in left.view(list)])

    val, = mapping.view(dict).values()
    # Some kind of runtime failure
    if val != '42':
        raise ValueError("Bad mapping")

    left_viz, = most_common_viz(left)
    right_viz, = most_common_viz(right)

    return mapping, left, right, left_viz, right_viz


def optional_artifact_pipeline(ctx, int_sequence, single_int=None):
    optional_artifact_method = ctx.get_action(
        'dummy_plugin', 'optional_artifacts_method')

    if single_int is None:
        # not a nested pipeline, just sharing the ctx object
        single_int = pointless_pipeline(ctx)

    num1 = single_int.view(int)
    ints, = optional_artifact_method(int_sequence, num1)
    return ints


def visualizer_only_pipeline(ctx, mapping):
    no_input_viz = ctx.get_action('dummy_plugin', 'no_input_viz')
    mapping_viz = ctx.get_action('dummy_plugin', 'mapping_viz')

    viz1, = no_input_viz()
    viz2, = mapping_viz(mapping, mapping, 'foo', 'bar')

    return viz1, viz2


def pipelines_in_pipeline(ctx, int_sequence, mapping):
    pointless_pipeline = ctx.get_action('dummy_plugin', 'pointless_pipeline')
    typical_pipeline = ctx.get_action('dummy_plugin', 'typical_pipeline')
    visualizer_only_pipeline = ctx.get_action(
        'dummy_plugin', 'visualizer_only_pipeline')

    results = []
    results += pointless_pipeline()
    typical_results = typical_pipeline(int_sequence, mapping, True)
    results += typical_results
    results += visualizer_only_pipeline(typical_results[0])

    return tuple(results)


def resumable_pipeline(ctx, int_list, int_dict, fail=False):
    """ This pipeline is designed to be called first with fail=True then a
    second time with fail=False. The second call is meant to reuse cached
    results from the first call
    """
    list_of_ints = ctx.get_action('dummy_plugin', 'list_of_ints')
    dict_of_ints = ctx.get_action('dummy_plugin', 'dict_of_ints')

    list_return, = list_of_ints(int_list)
    dict_return, = dict_of_ints(int_dict)

    if fail:
        list_uuids = [str(result.uuid) for result in list_return.values()]
        dict_uuids = [str(result.uuid) for result in dict_return.values()]

        raise ValueError(f'{list_uuids}_{dict_uuids}')

    return list_return, dict_return


# Either both int1 and string should be default or neither should be
def resumable_varied_pipeline(ctx, ints1, ints2, metadata, int1=None,
                              string='None', fail=False):
    varied_method = ctx.get_action('dummy_plugin', 'varied_method')
    list_of_ints = ctx.get_action('dummy_plugin', 'list_of_ints')
    dict_of_ints = ctx.get_action('dummy_plugin', 'dict_of_ints')
    identity_with_metadata = ctx.get_action('dummy_plugin',
                                            'identity_with_metadata')
    most_common_viz = ctx.get_action('dummy_plugin', 'most_common_viz')

    if int1 is None and string == 'None':
        ints1_ret, ints2_ret, int1_ret = varied_method(ints1, ints2)
    else:
        ints1_ret, ints2_ret, int1_ret = varied_method(
            ints1, ints2, int1, string)

    list_ret, = list_of_ints(ints1_ret)
    dict_ret, = dict_of_ints(ints1)

    identity_ret, = identity_with_metadata(ints2[0], metadata)

    viz_ret, = most_common_viz(ints2[1])

    if fail:
        uuids = []

        uuids.append([str(result.uuid) for result in ints1_ret.values()])
        uuids.append([str(result.uuid) for result in ints2_ret.values()])
        uuids.append(str(int1_ret.uuid))

        uuids.append([str(result.uuid) for result in list_ret.values()])
        uuids.append([str(result.uuid) for result in dict_ret.values()])

        uuids.append(str(identity_ret.uuid))

        uuids.append(str(viz_ret.uuid))

        raise PipelineError(uuids)

    return (ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, identity_ret,
            viz_ret)


# Either both int1 and string should be default or neither should be
def resumable_nested_varied_pipeline(ctx, ints1, ints2, metadata, int1=None,
                                     string='None', fail=False):
    internal_pipeline = ctx.get_action('dummy_plugin',
                                       'internal_fail_pipeline')
    list_of_ints = ctx.get_action('dummy_plugin', 'list_of_ints')
    dict_of_ints = ctx.get_action('dummy_plugin', 'dict_of_ints')
    identity_with_metadata = ctx.get_action('dummy_plugin',
                                            'identity_with_metadata')
    most_common_viz = ctx.get_action('dummy_plugin', 'most_common_viz')

    list_ret, = list_of_ints(ints1)
    dict_ret, = dict_of_ints(ints1)

    identity_ret, = identity_with_metadata(ints2[0], metadata)

    viz_ret, = most_common_viz(ints2[1])

    try:
        if int1 is None and string == 'None':
            ints1_ret, ints2_ret, int1_ret = internal_pipeline(
                ints1, ints2, fail=fail)._result()
        else:
            ints1_ret, ints2_ret, int1_ret = internal_pipeline(
                ints1, ints2, int1, string, fail=fail)._result()
    except PipelineError as e:
        uuids = [uuid for uuid in e.uuids]

        uuids.append([str(result.uuid) for result in list_ret.values()])
        uuids.append([str(result.uuid) for result in dict_ret.values()])

        uuids.append(str(identity_ret.uuid))

        uuids.append(str(viz_ret.uuid))

        raise PipelineError(uuids)

    return (ints1_ret, ints2_ret, int1_ret, list_ret, dict_ret, identity_ret,
            viz_ret)


# Either both int1 and string should be default or neither should be
def internal_fail_pipeline(ctx, ints1, ints2, int1=None, string='None',
                           fail=False):
    varied_method = ctx.get_action('dummy_plugin', 'varied_method')

    if int1 is None and string == 'None':
        ints1_ret, ints2_ret, int1_ret = varied_method(
            ints1, ints2)
    else:
        ints1_ret, ints2_ret, int1_ret = varied_method(
            ints1, ints2, int1, string)

    if fail:
        uuids = []

        uuids.append([str(result.uuid) for result in ints1_ret.values()])
        uuids.append([str(result.uuid) for result in ints2_ret.values()])
        uuids.append(str(int1_ret.uuid))

        raise PipelineError(uuids)

    return ints1_ret, ints2_ret, int1_ret


def de_facto_list_pipeline(ctx, kwarg=False, non_proxies=False):
    returns_int = ctx.get_action('dummy_plugin', 'returns_int')
    list_of_ints = ctx.get_action('dummy_plugin', 'list_of_ints')
    num_ints = 3

    ints = []
    for i in range(num_ints):
        ints_ret, = returns_int(i)
        ints.append(ints_ret)

    if non_proxies:
        ints.append(ctx.make_artifact(SingleInt, num_ints + 1))

    if kwarg:
        ret, = list_of_ints(ints=ints)
    else:
        ret, = list_of_ints(ints)

    return ret


def de_facto_dict_pipeline(ctx, kwarg=False, non_proxies=False):
    returns_int = ctx.get_action('dummy_plugin', 'returns_int')
    dict_of_ints = ctx.get_action('dummy_plugin', 'dict_of_ints')
    num_ints = 3

    ints = {}
    for i in range(num_ints):
        ints_ret, = returns_int(i)
        ints[str(i + 1)] = ints_ret

    if non_proxies:
        ints[str(num_ints + 2)] = ctx.make_artifact(SingleInt, num_ints + 1)

    if kwarg:
        ret, = dict_of_ints(ints=ints)
    else:
        ret, = dict_of_ints(ints)

    return ret


def list_pipeline(ctx, ints):
    assert isinstance(ints, list)
    return ([ctx.make_artifact(SingleInt, 4),
             ctx.make_artifact(SingleInt, 5)])


def collection_pipeline(ctx, ints):
    assert isinstance(ints, ResultCollection)
    return {'key1': ctx.make_artifact(SingleInt, 4),
            'key2': ctx.make_artifact(SingleInt, 5)}


def de_facto_collection_pipeline(ctx):
    method = ctx.get_action('dummy_plugin', 'no_input_method')

    art1, = method()
    art2, = method()

    return [art1, art2]


def pointless_pipeline(ctx):
    # Use a real type expression instead of a string.
    return ctx.make_artifact(SingleInt, 4)


def failing_pipeline(ctx, int_sequence, break_from='arity'):
    merge_mappings = ctx.get_action('dummy_plugin', 'merge_mappings')

    list_ = int_sequence.view(list)
    if list_:
        integer = list_[0]
    else:
        integer = 0

    # Made here so that we can make sure it gets cleaned up
    wrong_output = ctx.make_artifact(SingleInt, integer)

    if break_from == 'arity':
        return int_sequence, int_sequence, int_sequence
    elif break_from == 'return-view':
        return None
    elif break_from == 'type':
        return wrong_output
    elif break_from == 'method':
        a = ctx.make_artifact(Mapping, {'foo': 'a'})
        b = ctx.make_artifact(Mapping, {'foo': 'b'})
        # has the same key
        merge_mappings(a, b)
    elif break_from == 'no-plugin':
        ctx.get_action('not%a$plugin', 'foo')
    elif break_from == 'no-action':
        ctx.get_action('dummy_plugin', 'not%a$method')
    else:
        raise ValueError('this never works')
